#!/usr/bin/python
# -*- coding:utf-8 -*- 

import sys, getopt
import pandas, psycopg2
import csv
#import statsmodels.formula.api as smf
from sklearn import preprocessing
import math
import time
from heapq import *
import logging
from capexplain.similarity.category_similarity_matrix import *
from capexplain.similarity.category_similarity_naive import *
from capexplain.similarity.category_network_embedding import *
from capexplain.utils import *
from capexplain.pattern_model.LocalRegressionPattern import *
from capexplain.cl.cfgoption import DictLike
from capexplain.explanation_model.explanation_model import *
#from build.lib.capexplain.database.dbaccess import DBConnection
# setup logging
# log = logging.getLogger(__name__)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s %(levelname)s %(name)s line %(lineno)d: %(message)s')
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)


TEST_ID = ''
#********************************************************************************
# Configuration for Explanation generation
class ExplConfig(DictLike):

    DEFAULT_RESULT_TABLE = 'pub_large_no_domain'
    # DEFAULT_RESULT_TABLE = 'crime_clean_100000_2'
    DEFAULT_PATTERN_TABLE = 'dev.pub'
    # DEFAULT_PATTERN_TABLE = 'dev.crime_clean_100000'
    DEFAULT_QUESTION_PATH = './input/user_question.csv'

    EXAMPLE_NETWORK_EMBEDDING_PATH = './input/NETWORK_EMBEDDING'
    EXAMPLE_SIMILARITY_MATRIX_PATH = './input/SIMILARITY_DEFINITION'
    DEFAULT_AGGREGATE_COLUMN = '*'
    DEFAULT_EPSILON = 0.25
    DEFAULT_LAMBDA = 0.5
    TOP_K = 10
    PARAMETER_DEV_WEIGHT = 1.0
    # global MATERIALIZED_CNT
    MATERIALIZED_CNT = 0
    # global MATERIALIZED_DICT
    MATERIALIZED_DICT = dict()
    # global VISITED_DICT
    VISITED_DICT = dict()

    REGRESSION_PACKAGES = [ 'scikit-learn', 'statsmodels' ]
    
    def __init__(self,
                 query_result_table = DEFAULT_RESULT_TABLE,
                 pattern_table = DEFAULT_PATTERN_TABLE,
                 user_question_file = DEFAULT_QUESTION_PATH,
                 outputfile = '',
                 aggregate_column = DEFAULT_AGGREGATE_COLUMN,
                 regression_package = 'statsmodels'
    ):
        self.pattern_table = pattern_table
        self.query_result_table = query_result_table
        self.user_question_file = user_question_file
        self.outputfile = outputfile
        self.aggregate_column = aggregate_column
        self.regression_package = regression_package
        self.global_patterns =None
        self.schema = None
        self.global_patterns_dict = None
        self.conn = self.cur = None
#         try:
#             self.conn = psycopg2.connect(dbname=DBConn.db,
#                                          user=DBConn.user,
#                                          host=DBConn.host,
#                                          port=DBConn.port,
#                                          password=DBConn.password)
#             self.pattern_table = DBConn.local_table[:-6]
#             self.cur = self.conn.cursor()
#         except psycopg2.OperationalError:
#             print('Fail to connect to the database!')


    def __str__(self):
        return self.__dict__.__str__()

class TopkHeap(object):
    def __init__(self, k):
        self.topk = k
        self.data = []
 
    def Push(self, elem):
        if len(self.data) < self.topk:
            heappush(self.data, elem)
        else:
            # topk_small = self.data[0]
            # if elem[0] > topk_small[0]:
            #     heapreplace(self.data, elem)
            topk_small = self.data[0]
            if elem.score > topk_small.score:
                heapreplace(self.data, elem)
    def MinValue(self):
        # return min(list(map(lambda x: x[0], self.data)))
        return min(list(map(lambda x: x.score, self.data)))
    def MaxValue(self):
        # return max(list(map(lambda x: x[0], self.data)))
        return max(list(map(lambda x: x.score, self.data)))
    def TopK(self):
        return [x for x in reversed([heappop(self.data) for x in range(len(self.data))])]
    def HeapSize(self):
        return len(self.data)


def predict(local_pattern, t):
    if local_pattern[4] == 'const':
        # predictY = float(local_pattern[7][1:-1])
        predictY = float(local_pattern[6][1:-1].split(',')[0])
    elif local_pattern[4] == 'linear':
        # print(local_pattern, t)
        v = get_V_value(local_pattern[2], t)
        if isinstance(local_pattern[7], str):
            params_str = local_pattern[7].split('\n')
            params_dict = {}
            for i in range(0, len(params_str)-1):
                p_cate = re.compile(r'(.*)\[T\.\s*(.*)\]\s+(-?\d+\.\d+)')
                cate_res = p_cate.findall(params_str[i])
                if len(cate_res) != 0:
                    cate_res = cate_res[0]
                    v_attr = cate_res[0]
                    v_val = cate_res[1]
                    param = float(cate_res[2])
                    if v_attr not in params_dict:
                        params_dict[v_attr] = {}
                    params_dict[v_attr][v_val] = param
                else:
                    p_nume = re.compile(r'([^\s]+)\s+(-?\d+\.\d+)')
                    nume_res = p_nume.findall(params_str[i])
                    if len(nume_res) == 0:
                        continue
                    nume_res = nume_res[0]
                    v_attr = nume_res[0]
                    param = float(nume_res[1])
                    params_dict[v_attr] = param
        else:
            params_dict = local_pattern[7]
        predictY = params_dict['Intercept']
        for v_attr in t:
            v_key = '{}[T.{}]'.format(v_attr, t[v_attr])
            if v_key in params_dict:
                predictY += params_dict[v_key]
            else:
                if v_attr in params_dict:
                    predictY += params_dict[v_attr] * float(t[v_attr])
            
    return predictY


        
def tuple_distance(t1, t2, var_attr, cat_sim, num_dis_norm, agg_col):
    """Compute the similarity between two tuples t1 and t2 on their attributes var_attr

    Args:
        t1, t2: two tuples
        var_attr: variable attributes
        cat_sim: the similarity measure for categorical attributes
        num_dis_norm: normalization terms for numerical attributes
        agg_col: the column of aggregated value
    Returns:
        the Gower similarity between t1 and t2
    """
    dis = 0.0
    cnt = 0
    if var_attr is None:
        var_attr = t1.keys()
    max_dis = 0.0

    for v_col in var_attr:
        col = v_col.replace(' ', '')
        
        if col not in t1 and col not in t2:
            if col == 'name':
                dis += 10000
                cnt += 1
            continue
        if col not in t1 or col not in t2:
            if col == 'name':
                dis += 10000
                cnt += 1
            #     if 1 > max_dis:
            #         max_dis = 1
            # else:
            #     dis += 1
            #     cnt += 1
            #     if 1 > max_dis:
            #         max_dis = 1
            continue
        if col == 'name':
            if t1[col] != t2[col]:
                dis += 10000
            cnt += 1
            continue

        if col == 'venue' or col == 'pubkey':
            if t1[col] != t2[col]:
                dis += 0.25
            cnt += 1
            continue

        if cat_sim.is_categorical(col):
            t1_key = str(t1[col]).replace("'", '').replace(' ', '')
            t2_key = str(t2[col]).replace("'", '').replace(' ', '')
            s = 0
            if t1[col] == t2[col]:
                s = 1
            else:
                if col == 'venue':
                    s = 0.5
            s = cat_sim.compute_similarity(col, t1_key, t2_key, agg_col)
            if s == 0:
                dis += 1
                max_dis = 1
            else:
                dis += (((1.0/s)) * ((1.0/s))) / 100
                # dis += (1-s) * (1-s)
                if math.sqrt((((1.0/s)) * ((1.0/s)) - 1) / 100) > max_dis:
                    max_dis = math.sqrt((((1.0/s)) * ((1.0/s)) - 1) / 100)
            cnt += 1
        else:
            # print( num_dis_norm[col])
            # if col not in num_dis_norm or num_dis_norm[col]['range'] is None:
            #     if t1[col] == t2[col]:
            #         dis += 0
            #     else:
            #         dis += 1
            # else:
            #     if col != agg_col and col != 'index':
            #         # temp = abs(t1[col] - t2[col]) / (num_dis_norm[col]['range'])
            #         temp = abs(t1[col] - t2[col])
            #         # dis += 0.5 * math.pow(temp, 4)
            #         dis += math.pow(temp, 8)
            #     cnt += 1

            if col not in num_dis_norm or num_dis_norm[col]['range'] is None:
                if t1[col] == t2[col]:
                    dis += 0
                else:
                    dis += 1
            else:
                if col != agg_col and col != 'index':
                    # temp = (t1[col] - t2[col]) * (t1[col] - t2[col]) / (num_dis_norm[col]['range'] * num_dis_norm[col]['range'])
                    # temp = abs(t1[col] - t2[col]) / (num_dis_norm[col]['range'])
                    if isinstance(t1[col], datetime.date):
                        # print(398, t1, t2)
                        # print(col, t1[col], t2[col])
                        diff = datetime.datetime(t1[col].year, t1[col].month, t1[col].day) - datetime.datetime.strptime(t2[col], "%Y-%m-%d")
                        temp = diff.days
                    else:
                        # print(398, t1, t2)
                        # print(col, t1[col], t2[col])
                        temp = abs(float(t1[col]) - float(t2[col]))
                    # if temp != 0:
                    #     temp = 1/(1+(math.exp(-temp+2)))

                    dis += 0.5 * math.pow(temp, 8)
                    if temp > max_dis:
                        max_dis = temp
                cnt += 1
            
    return math.pow(dis, 0.5)
    # return max_dis

def get_local_patterns(F, Fv, V, agg_col, model_type, t, conn, cur, pat_table_name, res_table_name):
    local_patterns = []
    local_patterns_dict = {}
        
    if model_type is not None:
        mt_predicate = " AND model='{}'".format(model_type)
    else:
        mt_predicate = ''
    if Fv is not None:
        local_pattern_query = '''SELECT * FROM {} WHERE array_to_string(fixed, ', ')='{}' AND 
            REPLACE(array_to_string(fixed_value, ', '), '"', '') LIKE '%{}%' AND 
            array_to_string(variable, ', ') = '{}' AND
            agg='{}'{};'''.format(
                pat_table_name + '_local' + TEST_ID, str(F).replace("\'", '').replace('[', '').replace(']', ''), 
                str(Fv).replace("\'", '').replace('[', '').replace(']', ''),
                str(V).replace("\'", '').replace('[', '').replace(']', ''), 
                agg_col, mt_predicate
            )
        # print(406, Fv, local_pattern_query)
    else:
        tF = get_F_value(F, t)
        # local_pattern_query = '''SELECT * FROM {} WHERE array_to_string(fixed, ', ')='{}' AND 
        #     REPLACE(array_to_string(fixed_value, ', '), '"', '') = REPLACE('{}', '"', '') AND array_to_string(variable, ', ')='{}' AND
        #     agg='{}'{};'''.format(
        #         pat_table_name + '_local' + TEST_ID, str(F).replace("\'", '').replace('[', '').replace(']', ''), 
        #         str(tF).replace("\'", '"').replace('[', '').replace(']', ''),
        #         str(V).replace("\'", '').replace('[', '').replace(']', ''), 
        #         agg_col, mt_predicate
        #     )
        local_pattern_query = '''SELECT * FROM {} WHERE array_to_string(fixed, ', ')='{}' AND
            REPLACE(array_to_string(fixed_value, ', '), '"', '') LIKE '%{}%' AND array_to_string(variable, ', ')='{}' AND
            agg='{}'{};'''.format(
                pat_table_name + '_local'+TEST_ID, str(F).replace("\'", '').replace('[', '').replace(']', ''),
                '%'.join(list(map(str, tF))),
                str(V).replace("\'", '').replace('[', '').replace(']', ''),
                agg_col, mt_predicate
            )
    # print(293, local_pattern_query)
    cur.execute(local_pattern_query)
    local_patterns = cur.fetchall()
        
    return local_patterns


def get_tuples_by_F(local_pattern, f_value, cur, table_name, cat_sim):
    def tuple_column_to_str_in_where_clause_2(col_value):
        # print(col_value, cat_sim.is_categorical(col_value[0]))
        if cat_sim.is_categorical(col_value[0]):
            # return "like '%" + str(col_value[1]) + "%'"
            return "= '" + str(col_value[1]) + "'"
        else:
            if is_float(col_value[1]):
                return '=' + str(col_value[1])
            else:
                # return "like '%" + str(col_value[1]) + "%'"
                return "= '" + str(col_value[1]) + "'"

    F = str(local_pattern[0]).replace("\'", '')[1:-1]
    V = str(local_pattern[2]).replace("\'", '')[1:-1]
    F_list = F.split(', ')
    V_list = V.split(', ')
    where_clause = ' AND '.join(list(map(lambda x: "{} {}".format(x[0], x[1]), zip(local_pattern[0], map(tuple_column_to_str_in_where_clause_2, zip(F_list, f_value))))))
    tuples_query = '''SELECT {},{},{} as {} FROM {} WHERE {} GROUP BY {}, {};'''.format(
            F, V, local_pattern[3].replace('_', '(') + ')', local_pattern[3], table_name, where_clause, F, V
        )
    # tuples_query = "SELECT * FROM {} WHERE {};".format(table_name, where_clause)
    
    # column_name_query = "SELECT column_name FROM information_schema.columns where table_name='{}';".format(table_name)
    # print(column_name_query)
    # cur.execute(column_name_query)
    # column_name = cur.fetchall()
    column_name = F_list + V_list + [local_pattern[3]]
    cur.execute(tuples_query)
    # print(tuples_query)
    tuples = []
    res = cur.fetchall()
    min_agg = 1e10
    max_agg = -1e10
    for row in res:
        min_agg = min(min_agg, row[-1])
        max_agg = max(max_agg, row[-1])
        tuples.append(dict(zip(map(lambda x: x, column_name), row)))
        # row_data = {}
        # cnt = 0 
        # for f_attr in F_list:    
        #     if is_float(row[cnt]):
        #         row_data[f_attr] = float(row[cnt])
        #     elif is_integer(row[cnt]):
        #         row_data[f_attr] = float(int(row[cnt]))
        #     else:
        #         row_data[f_attr] = t[cnt]
        #     cnt += 1
        # for v_attr in V_list:
        #     if is_float(row[cnt]):
        #         row_data[v_attr] = float(row[cnt])
        #     elif is_integer(row[cnt]):
        #         row_data[v_attr] = float(int(row[cnt]))
        #     else:
        #         row_data[v_attr] = row[cnt]
        #     cnt += 1
        # row_data[pat[4] + '(' + pat[3] + ')'] = row[-1]
        # tuples.append(row_data)
    return tuples, max_agg - min_agg

def get_tuples_by_F_V2(lp1, lp2, f_value, v_value, conn, cur, table_name, cat_sim):
    def tuple_column_to_str_in_where_clause_2(col_value):
        # print(col_value, cat_sim.is_categorical(col_value[0]))
        if cat_sim.is_categorical(col_value[0]):
            # return "like '%" + str(col_value[1]) + "%'"
            return "= '" + str(col_value[1]) + "'"
        else:
            if is_float(col_value[1]):
                return '=' + str(col_value[1])
            else:
                # return "like '%" + str(col_value[1]) + "%'"
                return "= '" + str(col_value[1]) + "'"

    F1 = str(lp1[0]).replace("\'", '')[1:-1]
    V1 = str(lp1[2]).replace("\'", '')[1:-1]
    F1_list = F1.split(', ')
    V1_list = V1.split(', ')
    # print(lp1)
    # print(lp2)
    F2 = str(lp2[0]).replace("\'", '')[1:-1]
    V2 = str(lp2[2]).replace("\'", '')[1:-1]
    F2_list = F2.split(', ')
    V2_list = V2.split(', ')
    where_clause = ' AND '.join(list(map(lambda x: "{} {}".format(x[0], x[1]), zip(lp1[0], map(tuple_column_to_str_in_where_clause_2, zip(F1_list, f_value))))))
    where_clause += ' AND ' 
    where_clause += ' AND '.join(list(map(lambda x: "{} {}".format(x[0], x[1]), zip(lp1[2], map(tuple_column_to_str_in_where_clause_2, zip(V1_list, v_value))))))
    tuples_query = '''SELECT {},{},{} as {} FROM {} WHERE {} GROUP BY {}, {};'''.format(
            F2, V2, lp2[3].replace('_', '(') + ')', lp2[3], table_name, where_clause, F2, V2
        )
    
    column_name = F2_list + V2_list + [lp2[3]]
    cur.execute(tuples_query)
    # print(tuples_query)
    tuples = []
    res = cur.fetchall()
    min_agg = 1e10
    max_agg = -1e10
    for row in res:
        min_agg = min(min_agg, row[-1])
        max_agg = max(max_agg, row[-1])
        tuples.append(dict(zip(map(lambda x: x, column_name), row)))
        
    return tuples, max_agg - min_agg


def get_tuples_by_F_V(lp1, lp2, f_value, v_value, conn, cur, table_name, cat_sim, ExplConfig):
    def tuple_column_to_str_in_where_clause_2(col_value):
        # print(col_value, cat_sim.is_categorical(col_value[0]))
        if cat_sim.is_categorical(col_value[0]):
            # return "like '%" + str(col_value[1]) + "%'"
            return "= '" + str(col_value[1]) + "'"
        else:
            if is_float(col_value[1]):
                return '=' + str(col_value[1])
            else:
                # return "like '%" + str(col_value[1]) + "%'"
                return "= '" + str(col_value[1]) + "'"
    def tuple_column_to_str_in_where_clause_3(col_value):
        # print(col_value, cat_sim.is_categorical(col_value[0]))
        if cat_sim.is_categorical(col_value[0]):
            # return "like '%" + str(col_value[1]) + "%'"
            return "= '" + str(col_value[1]) + "'"
        else:
            if is_float(col_value[1]):
                return '>=' + str(col_value[1])
            else:
                # return "like '%" + str(col_value[1]) + "%'"
                return "= '" + str(col_value[1]) + "'"

    def tuple_column_to_str_in_where_clause_4(col_value):
        # print(col_value, cat_sim.is_categorical(col_value[0]))
        if cat_sim.is_categorical(col_value[0]):
            # return "like '%" + str(col_value[1]) + "%'"
            return "= '" + str(col_value[1]) + "'"
        else:
            if is_float(col_value[1]):
                return '<=' + str(col_value[1])
            else:
                # return "like '%" + str(col_value[1]) + "%'"
                return "= '" + str(col_value[1]) + "'"

    F1 = str(lp1[0]).replace("\'", '')[1:-1]
    V1 = str(lp1[2]).replace("\'", '')[1:-1]
    F1_list = F1.split(', ')
    V1_list = V1.split(', ')

    F2 = str(lp2[0]).replace("\'", '')[1:-1]
    V2 = str(lp2[2]).replace("\'", '')[1:-1]
    F2_list = F2.split(', ')
    V2_list = V2.split(', ')
    G_list = sorted(F2_list + V2_list)
    G_key = str(G_list).replace("\'", '')[1:-1]
    f_value_key = str(f_value).replace("\'", '')[1:-1]
    if lp2[3] == 'count':
        agg_fun = 'count(*)'
    else:
        agg_fun = lp2[3].replace('_', '(') + ')'



    if G_key not in ExplConfig.MATERIALIZED_DICT:
        ExplConfig.MATERIALIZED_DICT[G_key] = dict()
        
    if f_value_key not in ExplConfig.MATERIALIZED_DICT[G_key]:
        ExplConfig.MATERIALIZED_DICT[G_key][f_value_key] = ExplConfig.MATERIALIZED_CNT
        dv_query = '''DROP VIEW IF EXISTS MV_{};'''.format(str(ExplConfig.MATERIALIZED_CNT))
        cur.execute(dv_query)
        # print(504, MATERIALIZED_CNT)
        cmv_query = '''
            CREATE VIEW MV_{} AS SELECT {}, {} as {} FROM {} WHERE {} GROUP BY {};
        '''.format(
            str(ExplConfig.MATERIALIZED_CNT), G_key, agg_fun, lp2[3], table_name,
            ' AND '.join(list(map(lambda x: "{} {}".format(x[0], x[1]), 
                zip(lp1[0], map(tuple_column_to_str_in_where_clause_2, zip(F1_list, f_value)))))),
            G_key
        )
        # print(585, cmv_query)
        cur.execute(cmv_query)
        conn.commit()
        ExplConfig.MATERIALIZED_CNT += 1


    where_clause = ' AND '.join(list(map(lambda x: "{} {}".format(x[0], x[1]), zip(lp1[0], map(tuple_column_to_str_in_where_clause_2, zip(F1_list, f_value))))))
    if v_value is not None:
        where_clause += ' AND ' 
        # print(593, v_value)
        v_range_l = map(lambda x: v_value[0][x] + v_value[1][x][0], range(len(v_value[0])))
        v_range_r = map(lambda x: v_value[0][x] + v_value[1][x][1], range(len(v_value[0])))
        where_clause += ' AND '.join(list(map(lambda x: "{} {}".format(x[0], x[1]), zip(lp1[2], 
            map(tuple_column_to_str_in_where_clause_3, zip(V1_list, v_range_l))))))
        where_clause += ' AND ' 
        where_clause += ' AND '.join(list(map(lambda x: "{} {}".format(x[0], x[1]), zip(lp1[2], 
            map(tuple_column_to_str_in_where_clause_4, zip(V1_list, v_range_r))))))
    tuples_query = '''SELECT {},{},{} FROM MV_{} WHERE {};'''.format(
            F2, V2, lp2[3], str(ExplConfig.MATERIALIZED_DICT[G_key][f_value_key]), where_clause
        )
    # print(604, tuples_query)
    column_name = F2_list + V2_list + [lp2[3]]
    cur.execute(tuples_query)
    # print(tuples_query)
    tuples = []
    tuples_dict = dict()
    res = cur.fetchall()
    min_agg = 1e10
    max_agg = -1e10
    for row in res:
        min_agg = min(min_agg, row[-1])
        max_agg = max(max_agg, row[-1])
        tuples.append(dict(zip(map(lambda x: x, column_name), row)))
        fv = get_F_value(lp2[0], tuples[-1])
        f_key = str(fv).replace('\'', '')[1:-1]
        if f_key not in tuples_dict:
            tuples_dict[f_key] = []
        tuples_dict[f_key].append(tuples[-1])
        
    return tuples, (min_agg, max_agg), tuples_dict


def get_tuples_by_gp_uq(gp, f_value, v_value, conn, cur, table_name, cat_sim, ExplConfig):
    def tuple_column_to_str_in_where_clause_2(col_value):
        # print(col_value, cat_sim.is_categorical(col_value[0]))
        if cat_sim.is_categorical(col_value[0]) or col_value[0] == 'year':
            # return "like '%" + str(col_value[1]) + "%'"
            return "= '" + str(col_value[1]) + "'"
        else:
            if is_float(col_value[1]):
                return '=' + str(col_value[1])
            else:
                # return "like '%" + str(col_value[1]) + "%'"
                return "= '" + str(col_value[1]) + "'"
    
    F1 = str(gp[0]).replace("\'", '')[1:-1]
    V1 = str(gp[1]).replace("\'", '')[1:-1]
    F1_list = F1.split(', ')
    V1_list = V1.split(', ')
    G_list = sorted(F1_list + V1_list)
    G_key = str(G_list).replace("\'", '')[1:-1]
    f_value_key = str(f_value).replace("\'", '')[1:-1]
    v_value_key = str(v_value).replace("\'", '')[1:-1]

    if gp[2] == 'count':
        agg_fun = 'count(*)'
    else:
        agg_fun = gp[2].replace('_', '(') + ')'

    
    if G_key not in ExplConfig.MATERIALIZED_DICT:
        ExplConfig.MATERIALIZED_DICT[G_key] = dict()
        
    if f_value_key not in ExplConfig.MATERIALIZED_DICT[G_key]:
        ExplConfig.MATERIALIZED_DICT[G_key][f_value_key] = ExplConfig.MATERIALIZED_CNT
        dv_query = '''DROP VIEW IF EXISTS MV_{};'''.format(str(ExplConfig.MATERIALIZED_CNT))
        cur.execute(dv_query)
        # print(504, MATERIALIZED_CNT)
        cmv_query = '''
            CREATE VIEW MV_{} AS SELECT {}, {} as {} FROM {} WHERE {} GROUP BY {};
        '''.format(
            str(ExplConfig.MATERIALIZED_CNT), G_key, agg_fun, gp[2], table_name,
            ' AND '.join(list(map(lambda x: "{} {}".format(x[0], x[1]), 
                zip(gp[0], map(tuple_column_to_str_in_where_clause_2, zip(F1_list, f_value)))))),
            G_key
        )
        # print(585, cmv_query)
        cur.execute(cmv_query)
        conn.commit()
        ExplConfig.MATERIALIZED_CNT += 1


    where_clause = ' AND '.join(
        list(map(lambda x: "{} {}".format(x[0], x[1]), zip(gp[0], map(
            tuple_column_to_str_in_where_clause_2, zip(F1_list, f_value)))))) + ' AND ' + \
        ' AND '.join(list(map(lambda x: "{} {}".format(x[0], x[1]), 
                zip(gp[1], map(tuple_column_to_str_in_where_clause_2, zip(V1_list, v_value))))))

    tuples_query = '''SELECT {} FROM MV_{} WHERE {};'''.format(
            gp[2], str(ExplConfig.MATERIALIZED_DICT[G_key][f_value_key]), where_clause
        )
    # print(530, tuples_query)
    cur.execute(tuples_query)
    res = cur.fetchall()
    return res[0]


def find_global_patterns_exact_match(global_patterns_dict, F_prime_set, V_set, agg_col, reg_type):
    gp_list = []
    for v_key in global_patterns_dict:
        # print(v_key, F_prime_set)
        F_key_set = set(f_key[1:-1].replace("'", '').split(', '))
        # print(545, f_key)
        # F_set = set(f_key)
        # print(579, f_key, F_set, t_set)
        if F_key_set != F_prime_set:
            continue
        
        for v_key in global_patterns_dict[f_key]:
            V_key_set = set(v_key[1:-1].replace("'", '').split(', '))
            if V_key_set != V_set:
                continue
            for pat in global_patterns_dict[f_key][v_key]:
                if pat[2] != agg_col or pat[3] != reg_type:
                    continue
                gp_list.append(pat)
    return gp_list

def find_patterns_refinement(global_patterns_dict, F_prime_set, V_set, agg_col, reg_type, ExplConfig):
    # pattern refinement can have different model types
    # e.g., JH’s #pub increases linearly, but JH’s #pub on VLDB remains a constant
    gp_list = []
    v_key = str(sorted(list(V_set)))
    for f_key in global_patterns_dict[0][v_key]:
        F_key_set = set(f_key[1:-1].replace("'", '').split(', '))
        if F_prime_set.issubset(F_key_set):
            for pat in global_patterns_dict[0][v_key][f_key]:
                if pat[2] == agg_col:
                    pat_key = f_key + '|,|' + v_key + '|,|' + pat[2] + '|,|' + pat[3]
                    
                    gp_list.append(pat)
                    if pat_key not in ExplConfig.VISITED_DICT:
                        ExplConfig.VISITED_DICT[pat_key] = True
    return gp_list


def score_of_explanation(t1, t2, cat_sim, num_dis_norm, dir, denominator = 1, lp1 = None, lp2 = None):
    if lp1 is None:
        return 1.0
    else:
        #print(lp1, lp2, t1, t2)
        agg_col = lp1[3]
        t1fv = dict(zip(lp1[0] + lp1[2], map(lambda x:x, get_F_value(lp1[0] + lp1[2], t1))))
        t2fv = dict(zip(lp2[0] + lp2[2], map(lambda x:x, get_F_value(lp2[0] + lp2[2], t2))))
        # if lp1 == lp2:
        #     t_sim = tuple_similarity(t1fv, t2fv, lp1[2], cat_sim, num_dis_norm, agg_col)
        # else:
        #     t_sim = tuple_similarity(t1fv, t2fv, None, cat_sim, num_dis_norm, agg_col)
        # t_sim = tuple_similarity(t1fv, t2fv, None, cat_sim, num_dis_norm, agg_col)
        t_dis = tuple_distance(t1fv, t2fv, None, cat_sim, num_dis_norm, agg_col)
        cnt1 = 0
        cnt2 = 0
        for a1 in t1:
            if a1 != 'lambda' and a1 != agg_col:
                cnt1 += 1
        for a2 in t2:
            if a2 != 'lambda' and a2 != agg_col:
                cnt2 += 1
        
        if len(t1.keys()) + 1 != len(t2.keys()):
            diff = len(t2.keys()) - len(t1.keys()) - 1
            if 'lambda' not in t2:
                diff += 1
            w = 1
            if 'name' in t2 and 'name' not in t1:
                w = 10000
            t_dis = math.sqrt(t_dis * t_dis + w * diff * diff)
            
            # print(488, t1, t2)
        # print local_cons[i].var_attr, row
        t1v = dict(zip(lp1[2], map(lambda x:x, get_V_value(lp1[2], t1))))        
        predicted_agg1 = predict(lp1, t1v)
        t2v = dict(zip(lp2[2], map(lambda x:x, get_V_value(lp2[2], t2))))        
        predicted_agg2 = predict(lp2, t2v)
        # deviation - counterbalance_needed 
        deviation = float(t1[agg_col]) - predicted_agg1
        counterbalance = t2[agg_col] - predicted_agg2
        # deviation_normalized = (t1[agg_col] - predicted_agg1) / predicted_agg1
        # influence = -deviation / counterbalance
        #score = (deviation + counterbalance) * t_sim
        #score = deviation * (math.exp(t_sim) - 1)
        

        # if t_sim == 1:
        #     score = deviation * -dir
        # else:
        #     score = deviation * math.exp(t_sim) * -dir
        if t_dis == 0:
            score = deviation * -dir
        else:
            score = deviation / t_dis * -dir
        
        # score = deviation * t_sim * -dir * 1 / (1 + math.exp(-(influence - 0.5)))
        # score *= math.log(deviation_normalized + 3)
        # score /= counterbalance
        # print(414, t1fv, t2fv)
        # print(t_dis, deviation, score)
        # return score / float(denominator) * cnt1 / cnt2
        return [100 * score / float(denominator), t_dis, 0, deviation, float(denominator)]

def compare_tuple(t1, t2):
    flag1 = True
    for a in t1:
        if (a != 'lambda' and a.find('_') == -1):
            if a not in t2:
                flag1 = False
            elif t1[a] != t2[a]:
                return 0
    flag2 = True
    for a in t2:
        if (a != 'lambda' and a.find('_') == -1):
            if a not in t1:
                flag2 = False
            elif t1[a] != t2[a]:
                return 0
    
    if flag1 and flag2:
        return -1
    elif flag1:
        return -1
    else:
        return 0

def DrillDown(global_patterns_dict, local_pattern, F_set, U_set, V_set, t_prime_coarser, t_coarser, t_prime, target_tuple,
        conn, cur, pat_table_name, res_table_name, cat_sim, num_dis_norm, 
        dir, query_result, norm_lb, dist_lb, tkheap, ExplConfig):
    reslist = []
    F_prime_set = F_set.union(U_set)
    agg_col = local_pattern[3]
    # gp2_list = find_global_patterns_exact_match(global_patterns_dict, F_prime_set, V_set, local_pattern[3], local_pattern[4])
    # gp2_list = find_patterns_refinement(global_patterns_dict, F_prime_set, V_set, local_pattern[3], local_pattern[4])
    gp2_list = find_patterns_refinement(global_patterns_dict, F_set, V_set, local_pattern[3], local_pattern[4], ExplConfig)
    # print(714, F_set, V_set, gp2_list)
    if len(gp2_list) == 0:
        return []
    for gp2 in gp2_list:
        # if random.random() < 0.3:
        #     continue
        # print(693, gp2)
        if dir == 1:
            dev_ub = abs(gp2[7])
        else:
            dev_ub = abs(gp2[6])
        # dev_ub = agg_maxmin[0][0] - agg_maxmin[0][1]
        k_score = tkheap.MinValue()
        # print(888, dev_ub, k_score, 100 * float(dev_ub) / (dist_lb * float(norm_lb)))
        if tkheap.HeapSize() == ExplConfig.TOP_K and 100 * float(dev_ub) / (dist_lb * float(norm_lb)) <= k_score:
            # print(890, dev_ub, dist_lb, norm_lb, gp2[0], gp2[1])
            continue
        # lp2_list = get_local_patterns(gp2[0], None, gp2[1], gp2[2], None, t_prime, conn, cur, pat_table_name, res_table_name)
        # print(731, t_prime)
        lp2_list = get_local_patterns(gp2[0], None, gp2[1], gp2[2], gp2[3], t_prime, conn, cur, pat_table_name, res_table_name)
        # print(733, lp2_list)
        if len(lp2_list) == 0:
            continue
        lp2 = lp2_list[0]
        
        f_value = get_F_value(local_pattern[0], t_prime)
        tuples_same_F, agg_range, tuples_same_F_dict = get_tuples_by_F_V(local_pattern, lp2, f_value, 
            # [get_V_value(local_pattern[2], t_prime), [[-3, 3]]],
            None,
            conn, cur, res_table_name, cat_sim,
            ExplConfig)
        lp3_list = get_local_patterns(lp2[0], f_value, lp2[2], lp2[3], lp2[4], t_prime, conn, cur, pat_table_name, res_table_name)
        # tuples_same_F, agg_range = get_tuples_by_F(local_pattern, lp2, f_value, 
        #     conn, cur, res_table_name, cat_sim)
        # tuples_same_F, agg_range, tuples_same_F_dict = get_tuples_by_F_V(local_pattern, lp2, f_value, None,
        #     conn, cur, res_table_name, cat_sim)
        # print(725, local_pattern[0], local_pattern[1], local_pattern[2], lp2[0], lp2[1], lp2[2])
        # print(725, len(tuples_same_F), len(lp3_list))
        for lp3 in lp3_list:
            if dir == 1:
                dev_ub = abs(lp3[9])
            else:
                dev_ub = abs(lp3[8])
            # dev_ub = agg_maxmin[0][0] - agg_maxmin[0][1]
            k_score = tkheap.MinValue()
            print(919, dev_ub, k_score, 100 * float(dev_ub) / (dist_lb * float(norm_lb)))
            if tkheap.HeapSize() == ExplConfig.TOP_K and 100 * float(dev_ub) / (dist_lb * float(norm_lb)) <= k_score:
                print(921, dev_ub, dist_lb, norm_lb, lp3[0], lp3[1])
                continue
            f_key = str(lp3[1]).replace('\'', '')[1:-1]
            if f_key in tuples_same_F_dict:
                for idx, row in enumerate(tuples_same_F_dict[f_key]):
                    for idx2, row2 in enumerate(t_coarser):
                        if get_V_value(local_pattern[2], row2) == get_V_value(local_pattern[2], row):
                            s = score_of_explanation(row, target_tuple, cat_sim, num_dis_norm, dir, float(row2[agg_col]), lp3, lp2)
                            break
                
                    # e.g. U = {Venue}, u = {ICDE}, do not need to check whether {Author, Venue} {Year} holds on (JH, ICDE)
                    # expected values are replaced with the average across year for all (JH, ICDE, year) tuples
                    #s = score_of_explanation(row, t_prime, cat_sim)
                    cmp_res = compare_tuple(row, target_tuple)
                    if cmp_res == 0: # row is not subset of target_tuple, target_tuple is not subset of row
                        # print(551, row, target_tuple, s)
                        # reslist.append([s[0], s[1:], dict(row), local_pattern, lp3, 1])  
                        reslist.append(Explanation(1, s[0], s[1], s[2], s[3], dir, dict(row), ExplConfig.TOP_K, local_pattern, lp3))
                    # else:
                    #     reslist.append([-s[0], s[1:], dict(row), local_pattern, lp3, 1])  
            # for f_key in tuples_same_F_dict:
            # # commented finer pattern holds only
            # if gp2 does not hold on (t[F], u):
            #     continue
            # e.g.{Author, venue} {Year} holds on (JH, ICDE)
            # lp3 = (F', (t[F], u), V, agg, a, m3)
            # s = score_of_explanation((t[F],u,t[V]), t, lp3, lp2)
            # f_prime_value = f_key.split(', ')
            # v_prime_value = get_V_value(lp2[0], row)
            # if v_prime_value[0]
            # f_prime_value = get_F_value(lp2[0], row)
            # lp3 = get_local_patterns(lp2[0], f_prime_value, lp2[2], lp2[3], lp2[4], row, conn, cur, pat_table_name, res_table_name)
            # lp3 = get_local_patterns(lp2[0], f_prime_value, lp2[2], lp2[3], lp2[4], tuples_same_F_dict[f_key][0], conn, cur, pat_table_name, res_table_name)

            # print(791, f_key, f_prime_value, len(lp3))
            # if len(lp3) == 0:
            #     continue
            # for row in tuples_same_F_dict[f_key]:
            #     s = score_of_explanation(row, target_tuple, cat_sim, num_dis_norm, dir, float(t_coarser[agg_col]), lp3[0], lp2)
            
            #     # e.g. U = {Venue}, u = {ICDE}, do not need to check whether {Author, Venue} {Year} holds on (JH, ICDE)
            #     # expected values are replaced with the average across year for all (JH, ICDE, year) tuples
            #     #s = score_of_explanation(row, t_prime, cat_sim)
            #     if not equal_tuple(row, target_tuple):
            #         # print(551, row, target_tuple, s)
            #         reslist.append([s[0], s[1:], dict(row), local_pattern, lp3[0], 1])

    
    return reslist

def find_explanation_regression_based(user_question_list, global_patterns, global_patterns_dict, 
    cat_sim, num_dis_norm, agg_col, conn, cur, pat_table_name, res_table_name, ExplConfig):

    """Find explanations for user questions

    Args:
        data: data['df'] is the data frame storing Q(R)
            data['le'] is the label encoder, data['ohe'] is the one-hot encoder
        user_question_list: list of user questions (t, dir), all questions have the same Q(R)
        cons: list of fixed attributes and variable attributes of global constraints
        cat_sim: the similarity measure for categorical attributes
        num_dis_norm: normalization terms for numerical attributes
        cons_epsilon: threshold for local regression constraints
        agg_col: the column of aggregated value
        regression_package: which package is used to compute regression 
    Returns:
        the top-k list of explanations for each user question
    """
    answer = [[] for i in range(len(user_question_list))]
    local_pattern_loading_time = 0
    question_validating_time = 0
    score_computing_time = 0
    score_computing_time_list = []
    result_merging_time = 0
    local_patterns_list = []
    # print(792, global_patterns)

    for j, uq in enumerate(user_question_list):
        dir = uq['dir']
        topK_heap = TopkHeap(ExplConfig.TOP_K)
        marked = {}

        t = dict(uq['target_tuple'])
        start = time.time()
        
        
        end = time.time()
        local_pattern_loading_time += end - start

        uq['global_patterns'] = find_patterns_relevant(
                global_patterns_dict, uq['target_tuple'], conn, cur, res_table_name, pat_table_name, cat_sim)

        # print(811, len(uq['global_patterns']))
        candidate_list = [[] for i in range(len(uq['global_patterns']))]
        top_k_lists = [[] for i in range(len(uq['global_patterns']))]
        validate_res_list = []
        # local_patterns_list.append(local_patterns)
        local_patterns = []
        
        psi = []
        
        ExplConfig.VISITED_DICT = dict()
        score_computing_time_cur_uq = 0
        score_computing_start = time.time()
        explanation_type = 0
        for i in range(0, len(uq['global_patterns'])):
            top_k_lists[i] = [4, uq['global_patterns'][i], t, []]
            local_patterns.append(None)
            F_key = str(sorted(uq['global_patterns'][i][0]))
            V_key = str(sorted(uq['global_patterns'][i][1]))
            pat_key = F_key + '|,|' + V_key + '|,|' + uq['global_patterns'][i][2] + '|,|' + uq['global_patterns'][i][3]
            if pat_key in ExplConfig.VISITED_DICT:
                continue
            ExplConfig.VISITED_DICT[pat_key] = True

            
            tF = get_F_value(uq['global_patterns'][i][0], t)
            local_pattern_query_fixed = '''SELECT * FROM {} 
                    WHERE array_to_string(fixed, ', ')='{}' AND 
                    array_to_string(fixed_value, ', ')='{}' AND
                    array_to_string(variable, ', ')='{}' AND 
                    agg='{}' AND model='{}'
                ORDER BY theta;
            '''.format(
                pat_table_name + '_local' + TEST_ID, 
                str(uq['global_patterns'][i][0]).replace("\'", '').replace('[', '').replace(']', ''),
                str(tF)[1:-1].replace("\'", ''),
                str(uq['global_patterns'][i][1]).replace("\'", '').replace('[', '').replace(']', ''), 
                uq['global_patterns'][i][2], uq['global_patterns'][i][3] 
            )
            cur.execute(local_pattern_query_fixed)
            res_fixed = cur.fetchall()
            if len(res_fixed) == 0:
                continue
            local_patterns[i] = res_fixed[0]
            # if len(local_patterns[i][2]) > 1 or local_patterns[i][2][0] != 'year':
            #     continue
            T_set = set(t.keys()).difference(set(['lambda', uq['global_patterns'][i][2]]))
            psi.append(0)
            agg_col = local_patterns[i][3]
            start = time.time()
            # print('PAT', i, local_patterns[i])
            # t_coarser_list, agg_range, t_coarser_dict = get_tuples_by_F_V(local_patterns[i], local_patterns[i], 
            #     get_F_value(local_patterns[i][0], t), [get_V_value(local_patterns[i][2], t), [[0, 0]]],
            #     conn, cur, res_table_name, cat_sim)
            t_t_list, agg_range, t_t_dict = get_tuples_by_F_V(local_patterns[i], local_patterns[i], 
                get_F_value(local_patterns[i][0], t), 
                #[get_V_value(local_patterns[i][2], t), [[-3, 3]]],
                None,
                conn, cur, res_table_name, cat_sim,
                ExplConfig)
            # print(t_t_list)
            dist_lb = 1e10
            dev_ub = 0
            for t_t in t_t_list:
                if compare_tuple(t_t, t) == 0:
                    s = score_of_explanation(t_t, t, cat_sim, num_dis_norm, dir, t_t[agg_col], local_patterns[i], local_patterns[i])
                    if str(t_t) not in marked:
                        marked[str(t_t)] = True
                        # topK_heap.Push([s[0], s[1:], 
                        #     list(map(lambda y: y[1], sorted(t_t.items(), key=lambda x: x[0]))), 
                        #     local_patterns[i], None, 0])
                        topK_heap.Push(Explanation(0, s[0], s[1], s[2], s[3], uq['dir'],
                            # list(map(lambda y: y[1], sorted(t_t.items(), key=lambda x: x[0]))), 
                            dict(t_t),
                            ExplConfig.TOP_K, local_patterns[i], None))
                    # top_k_lists[i][-1].append([s[0], s[1:], dict(t_t), local_patterns[i], None, 0])
                    top_k_lists[i][-1].append(Explanation(0, s[0], s[1], s[2], s[3], uq['dir'],
                            dict(t_t),
                            ExplConfig.TOP_K, local_patterns[i], None))
                    if s[1] < dist_lb:
                        dist_lb = s[1]
                    if abs(s[3]) > dev_ub:
                        dev_ub = abs(s[3])
            if dist_lb < 1e-10:
                dist_lb = 0.01

            
            end = time.time()
            question_validating_time += end - start

            
            F_set = set(local_patterns[i][0])
            V_set = set(local_patterns[i][2])
            
            # if 'venue' in local_patterns[i][0] and 'name' not in local_patterns[i][0]:
            #     continue
            # if 'pubkey' in local_patterns[i][0] and 'name' not in local_patterns[i][0]:
            #     continue
            # F union V \subsetneq G            
            
            # print(t_coarser)
            t_coarser_copy = list(t_t_list)
            # t_coarser_copy[agg_col] = 1
            # if len(local_patterns[i][0]) + len(local_patterns[i][2]) < len(t.keys()) -2:
                # if 'pubkey' in local_patterns[i][0] and 'name' not in local_patterns[i][0]:
                #     continue
            # if 'venue' in local_patterns[i][0] and 'name' not in local_patterns[i][0]:
            #     continue

            # dev_ub_list = get_dev_upper_bound(local_patterns[i])
            # dev_ub = agg_range[1] - agg_range[0]
            norm_lb = min(list(map(lambda x: x[agg_col], t_coarser_copy)))
            # cur_top_k_res = topK_heap.TopK()
            # k_score = min(map(lambda x:x[0], cur_top_k_res))
            k_score = topK_heap.MinValue()
            # k_score = topK_heap.MaxValue()
            print(1124, k_score, 100 * float(dev_ub) / (dist_lb * float(norm_lb)))
            if topK_heap.HeapSize() == ExplConfig.TOP_K and 100 * float(dev_ub) / (dist_lb * float(norm_lb)) <= k_score:
                print(1126, dev_ub, dist_lb, norm_lb, local_patterns[i][0], local_patterns[i][1])
                continue
            top_k_lists[i][-1] += DrillDown(global_patterns_dict, local_patterns[i], 
                F_set, T_set.difference(F_set.union(V_set)), V_set, t_coarser_copy, t_coarser_copy, t, t,
                conn, cur, pat_table_name, res_table_name, cat_sim, num_dis_norm,
                dir, uq['query_result'],
                norm_lb, dist_lb, topK_heap, ExplConfig)
            for tk in top_k_lists[i][-1]:
                # print(937, tk.to_string())
                # if str(tk[2]) not in marked:
                if str(tk.tuple_value) not in marked:
                    # marked[str(tk[2])] = True
                    marked[str(tk.tuple_value)] = True
                    # topK_heap.Push([tk[0], tk[1], 
                    #     list(map(lambda y: y[1], sorted(tk[2].items(), key=lambda x: x[0]))), 
                    #     tk[3], tk[4], tk[5]])
                    topK_heap.Push(tk)
            

            end = time.time()
            
            score_computing_time += end - start
                
                
        score_computing_end = time.time()             
        score_computing_time_cur_uq = score_computing_end - score_computing_start
        # uses heapq to manipulate merge of explanations from multiple constraints
        start = time.time()
        # complementary_explanation_list = []
        # heapify(complementary_explanation_list)
        # refute_explanation_list = []
        # for i in range(len(local_patterns)):
        #     if len(top_k_lists[i]) > 3 and len(top_k_lists[i][3]) > 0:
        #         for idx, tk in enumerate(top_k_lists[i][3]):
        #             top_k_lists[i][3][idx][0] = -tk[0]
        #             top_k_lists[i][3][idx][2] = list(map(lambda y: y[1], sorted(tk[2].items(), key=lambda x: x[0])))
        #         print(top_k_lists[i])
        #         heapify(top_k_lists[i][-1])
        #         # print(top_k_lists[i])
        #         poped_tuple = list(heappop(top_k_lists[i][-1]))
        #         poped_tuple.append(i)
        #         heappush(complementary_explanation_list, poped_tuple)
        #     else:
        #         refute_explanation_list.append(top_k_lists[i])
        answer[j] = [{} for i in range(ExplConfig.TOP_K)]

        # cnt = 0
        # while cnt < TOP_K and len(complementary_explanation_list) > 0:
        #     poped_tuple = heappop(complementary_explanation_list)
        #     if str(poped_tuple[2]) not in marked:
        #         marked[str(poped_tuple[2])] = True
        #         answer[j][cnt] = (poped_tuple[0], poped_tuple[1], poped_tuple[2], poped_tuple[3], poped_tuple[4], poped_tuple[5])
        #         cnt += 1
        #     if len(top_k_lists[poped_tuple[-1]][3]) == 0:
        #         for i in range(len(local_patterns)):
        #             if len(top_k_lists[i]) > 3 and len(top_k_lists[i][-1]) > 0:
        #                 poped_tuple2 = list(heappop(top_k_lists[i][-1]))
        #                 poped_tuple2.append(i)
        #                 heappush(complementary_explanation_list, poped_tuple2)
                        
        #         heapify(complementary_explanation_list)
        #     else:
        #         # for tk in top_k_lists[poped_tuple[3]][3]:
        #         #     print(tk)

        #         poped_tuple2 = list(heappop(top_k_lists[poped_tuple[-1]][-1]))
        #         poped_tuple2.append(poped_tuple[-1])
        #         heappush(complementary_explanation_list, poped_tuple2)
        answer[j] = topK_heap.TopK()
        end = time.time()
        result_merging_time += end - start
        score_computing_time_list.append([t, score_computing_time_cur_uq])
        # print(749, answer[j])
        # print(topK_heap)

        # for i in range(len(local_patterns)):
        #     # print("TOP K: ", top_k_lists[i])
        #     if len(top_k_lists[i]) > 3:
        #         print(top_k_lists[i])
        #         answer[j].append(sorted(top_k_lists[i][-1], key=lambda x: x[1], reverse=True)[0:TOP_K])
        #     else:
        #         answer[j].append(top_k_lists[i])
        #     # print(len(answer[j][-1]))



    print('Local pattern loading time: ' + str(local_pattern_loading_time) + 'seconds')
    print('Question validating time: ' + str(question_validating_time) + 'seconds')
    print('Score computing time: ' + str(score_computing_time) + 'seconds')
    print('Result merging time: ' + str(result_merging_time) + 'seconds')
    return answer, local_patterns_list, score_computing_time_list

def load_query_result(t, cur, query_result_table, agg_col):
    agg_arr = agg_col.split('_')
    agg = agg_arr[0]
    a = agg_arr[1]
    group_arr = list(t.keys())
    group_arr.remove('lambda')
    group_arr.remove('direction')
    group_arr.remove('agg_col')
    group = ', '.join(group_arr)
    aggregate_query = 'SELECT {}, {}({}) as {} FROM {} GROUP BY {};'.format(
            group, agg, a, agg_col, query_result_table, group
        )
    cur.execute(aggregate_query)
    res = cur.fetchall()
    qr = list(map(lambda y: dict(zip(map(lambda x: x, column_name), y)), res))
    
    return qr

def find_patterns_relevant(global_patterns_dict, t, conn, cur, query_table_name, pattern_table_name, cat_sim):
    
    g_pat_list = []
    l_pat_list = []
    res_list = []
    t_set = set(t.keys())
    # print(1036, global_patterns_dict[0].keys())
    for v_key in global_patterns_dict[0]:
        # print(pat, pat[0])
        V_set = set(v_key[1:-1].replace("'", '').split(', '))
        # print(545, f_key)
        # F_set = set(f_key)
        # print(579, v_key, V_set, t_set)
        if not V_set.issubset(t_set):
            continue
        
        for f_key in global_patterns_dict[0][v_key]:
            for pat in global_patterns_dict[0][v_key][f_key]:
                # F_set = set(f_key.split(','))
                F_set = set(f_key[1:-1].replace("'", '').split(', '))
                # print(587, f_key, F_set, V_set, t_set)
                if not F_set.issubset(t_set):
                    continue
                # print(1053, pat, t)
                if pat[2] not in t and pat[2] + '_star' not in t:
                    continue

                agg_value = get_tuples_by_gp_uq(pat, get_F_value(pat[0], t), get_V_value(pat[1], t), 
                    conn, cur, query_table_name, cat_sim,
                    ExplConfig)
                res_list.append([pat, agg_value[0]])

                
    res_list = sorted(res_list, key = lambda x: (len(x[0][0]) + len(x[0][1]), x[1]))
    g_pat_list = list(map(lambda x: x[0], res_list))
    # print(1065, g_pat_list)
    return g_pat_list


def load_user_question_from_file(global_patterns, global_patterns_dict, uq_path, schema=None, conn=None, cur=None, pattern_table='', query_result_table='', pf=None, cat_sim=None):
    '''
        load user questions
    '''
    uq = []
    with open(uq_path, 'rt') as uqfile:
        reader = csv.DictReader(uqfile, quotechar='\'')
        headers = reader.fieldnames
        #temp_data = csv.reader(uqfile, delimiter=',', quotechar='\'')
        #for row in temp_data:
        for row in reader:
            row_data = {}
            raw_row_data = {}
            agg_col = None
            for k, v in enumerate(headers):
                # print(k, v)
                if schema is None or v not in schema:
                    if v != 'direction':
                        if is_float(row[v]):
                            row_data[v] = float(row[v])
                        elif is_integer(row[v]):
                            row_data[v] = float(long(row[v]))
                        else:
                            row_data[v] = row[v]
                    if v not in schema and v != 'lambda' and v != 'direction':
                        agg_col = v
                else:
                    if row[v] != '*':
                        if v.startswith('count_') or v.startswith('sum_'):
                            agg_col = v 
                        if schema[v] == 'integer':
                            row_data[v] = int(row[v])
                            raw_row_data[v] = int(row[v])
                        elif schema[v].startswith('double') or schema[v].startswith('float'):
                            row_data[v] = float(row[v])
                            raw_row_data[v] = float(row[v])
                        else:    
                            row_data[v] = row[v]
                            raw_row_data[v] = row[v]

            if row['direction'][0] == 'h':
                dir = 1
            else:
                dir = -1
            # print(615, schema, raw_row_data, row_data)
            uq.append({'target_tuple': row_data, 'dir':dir})
            # uq[-1]['global_patterns'], uq[-1]['local_patterns'] = find_patterns_relevant(
            #     global_patterns_dict, uq[-1]['target_tuple'], cur, pattern_table
            # )
            # uq[-1]['global_patterns'] = find_patterns_relevant(
            #     global_patterns_dict, uq[-1]['target_tuple'], conn, cur, query_result_table, pattern_table, cat_sim)

            
            # print(739, j, g_pat_cnt_by_theta[j])
            # print(list(map(lambda x: str(x[0]) + ' ' + str(x[1]) + '     ', g_pat_list_by_theta[j][0])))
            # print(list(map(lambda x: str(x[0]) + ' ' + str(x[1]) + '     ' if len(x) > 1 else '   empty   ', g_pat_list_by_theta[j][1])))
    
            # uq[-1]['query_result'] = load_query_result(uq[-1]['target_tuple'], cur, query_result_table, agg_col)
            uq[-1]['query_result'] = []
    # print(992, uq[-1])
    return uq, global_patterns, global_patterns_dict


def load_patterns(cur, pat_table_name, query_table_name):
    '''
        load pre-defined constraints(currently only fixed attributes and variable attributes)
    '''
    global_pattern_table = pat_table_name + '_global'
    load_query = "SELECT * FROM {};".format(global_pattern_table)
    # load_query = "SELECT * FROM {} WHERE fixed = '[primary_type]' AND variable = '[community_area, arrest, week]';".format(global_pattern_table)
    
    cur.execute(load_query)
    res = cur.fetchall()
    patterns = []
    pattern_dict = [{}, {}]
    for pat in res:
        if 'name' in pat[1] or 'venue' in pat[1]:
            continue
        patterns.append(list(pat))
        # print(pat)
        # print(patterns[-1])
        # patterns[-1][0] = patterns[-1][0][1:-1].replace(' ', '').split(',')
        # patterns[-1][1] = patterns[-1][1][1:-1].replace(' ', '').split(',')
        f_key = str(sorted(patterns[-1][0]))
        # f_key = str(patterns[-1][0])
        # print(694, f_key)
        v_key = str(sorted(patterns[-1][1]))
        # v_key = str(patterns[-1][1])
        if v_key not in pattern_dict[0]:
            pattern_dict[0][v_key] = {}
        if f_key not in pattern_dict[0][v_key]:
            pattern_dict[0][v_key][f_key] = []
        pattern_dict[0][v_key][f_key].append(patterns[-1])
        if f_key not in pattern_dict[1]:
            pattern_dict[1][f_key] = {}
        if v_key not in pattern_dict[1][f_key]:
            pattern_dict[1][f_key][v_key] = []
        pattern_dict[1][f_key][v_key].append(patterns[-1])
    schema_query = '''select column_name, data_type, character_maximum_length
        from INFORMATION_SCHEMA.COLUMNS where table_name=\'{}\'
    '''.format(query_table_name);
    cur.execute(schema_query)
    res = cur.fetchall()
    schema = {}
    for s in res:
        schema[s[0]] = s[1]
    
    # logger.debug('patterns')
    # logger.debug(patterns)
    return patterns, schema, pattern_dict

class ExplanationGenerator:

    def __init__(self, config : ExplConfig, user_input_config = None):
        self.config = config 
        if user_input_config is not None:
            if 'pattern_table' in user_input_config:
                self.config.pattern_table = user_input_config['pattern_table']
            if 'query_result_table' in user_input_config:
                self.config.query_result_table = user_input_config['query_result_table']
            if 'user_question_file' in user_input_config:
                self.config.user_question_file = user_input_config['user_question_file']
            if 'outputfile' in user_input_config:
                self.config.outputfile = user_input_config['outputfile']
            if 'aggregate_column' in user_input_config:
                self.config.aggregate_column = user_input_config['aggregate_column']
                

    def initialize(self):
        ecf = self.config
        query_result_table = ecf.query_result_table
        pattern_table = ecf.pattern_table
        user_question_file = ecf.DEFAULT_QUESTION_PATH
        outputfile = ''
        aggregate_column = ecf.aggregate_column
        conn = ecf.conn
        cur = ecf.cur
        logger.debug(ecf)
        logger.debug("pattern_table is")
        logger.debug(pattern_table)

        logger.debug("query_result_table is")
        logger.debug(query_result_table)

        # print(opts)
        start = time.clock()
        logger.info("start explaining ...")
        self.global_patterns, self.schema, self.global_patterns_dict = load_patterns(cur, pattern_table, query_result_table)
        logger.debug("loaded patterns from database")

        if query_result_table.find('crime') == -1:
            self.category_similarity = CategorySimilarityNaive(cur=cur, table_name=query_result_table)
        else:
            self.category_similarity = CategorySimilarityNaive(cur=cur, table_name=query_result_table, embedding_table_list=[('community_area', 'community_area_loc')])
        # category_similarity = CategoryNetworkEmbedding(EXAMPLE_NETWORK_EMBEDDING_PATH, data['df'])
        #num_dis_norm = normalize_numerical_distance(data['df'])
        self.num_dis_norm = normalize_numerical_distance(cur=cur, table_name=query_result_table)
        end = time.clock()
        print('Loading time: ' + str(end-start) + 'seconds')


    def wrap_user_question(self, global_patterns, global_patterns_dict, uq_tuple, schema=None):
        '''
            wrap user questions
        '''
        row_data = {}
        agg_col = None
        if 'lambda' not in uq_tuple:
            uq_tuple['lambda'] = 0.1
        for k, v in enumerate(uq_tuple):
            # print(k, v)
            if schema is None or v not in schema:
                if v != 'direction':
                    if is_float(uq_tuple[v]):
                        row_data[v] = float(uq_tuple[v])
                    elif is_integer(uq_tuple[v]):
                        row_data[v] = float((uq_tuple[v]))
                    else:
                        row_data[v] = uq_tuple[v]
                if v not in schema and v != 'lambda' and v != 'direction':
                    agg_col = v
            else:
                if uq_tuple[v] != '*':
                    if v.startswith('count_') or v.startswith('sum_'):
                        agg_col = v 
                    if schema[v] == 'integer':
                        row_data[v] = int(uq_tuple[v])
                    elif schema[v].startswith('double') or schema[v].startswith('float'):
                        row_data[v] = float(uq_tuple[v])
                    else:    
                        row_data[v] = uq_tuple[v]

        if uq_tuple['direction'][0] == 'h':
            dir = 1
        else:
            dir = -1

        if 'count_*' in row_data:
            row_data['count'] = row_data['count_*']
        uq = {'target_tuple': row_data, 'dir':dir, 'query_result': []}

        logger.debug("uq is")
        logger.debug(uq)

        return [uq]


    def do_explain_online(self, uq_tuple):

        ecf = self.config
        query_result_table = ecf.query_result_table
        pattern_table = ecf.pattern_table
        outputfile = ''
        aggregate_column = ecf.aggregate_column
        conn = ecf.conn
        cur = ecf.cur

        logger.debug("pattern_table is")
        logger.debug(pattern_table)

        logger.debug("query_result_table is")
        logger.debug(query_result_table)
        
        Q = self.wrap_user_question(self.global_patterns, self.global_patterns_dict, uq_tuple, self.schema)

        logger.debug("start finding explanations ...")
        
        start = time.clock()
        #regression_package = 'scikit-learn'
        regression_package = 'statsmodels'

        explanations_list, local_patterns_list, score_computing_time_list = find_explanation_regression_based(
            Q, self.global_patterns, self.global_patterns_dict, self.category_similarity, self.num_dis_norm,
            aggregate_column, conn, cur, 
            pattern_table, query_result_table, ecf
        )
            
        end = time.clock()
        # print('Total querying time: ' + str(end-start) + 'seconds')
        logger.debug("finding explanations ... DONE")

        # for g_key in ecf.MATERIALIZED_DICT:
        #     for fv_key in ecf.MATERIALIZED_DICT[g_key]:
        #         dv_query = '''DROP VIEW IF EXISTS MV_{};'''.format(str(ecf.MATERIALIZED_DICT[g_key][fv_key]))
        #         cur.execute(dv_query)
        #         conn.commit()
        return explanations_list[0]


    def doExplain(self):
        # c=self.config
        # logger.info("start explaining ...")
        # start = time.clock()
        # data = load_data(c.query_result_file)
        # logger.debug("loaded query results from file")
        # constraints = load_constraints(ExplConfig.DEFAULT_CONSTRAINT_PATH)
        # logger.debug("loaded patterns from file")
        # Q = load_user_question(c.user_question_file)
        # logger.debug("loaded user question from file")
        # category_similarity = CategorySimilarityMatrix(ExplConfig.EXAMPLE_SIMILARITY_MATRIX_PATH)
        # #category_similarity = CategoryNetworkEmbedding(EXAMPLE_NETWORK_EMBEDDING_PATH, data['df'])
        # num_dis_norm = normalize_numerical_distance(data['df'])
        # end = time.clock()
        # logger.debug("done loading")
        # print('Loading time: ' + str(end-start) + 'seconds')

        # logger.debug("start finding explanations ...")
        # start = time.clock()
        # #regression_package = 'scikit-learn'
        # regression_package = 'statsmodels'
        # explanations_list = find_explanation_regression_based(data, Q, constraints, category_similarity, 
        #                                                       num_dis_norm, c.constraint_epsilon, 
        #                                                       aggregate_column, regression_package)
        # end = time.clock()
        # print('Total querying time: ' + str(end-start) + 'seconds')
        # logger.debug("finding explanations ... DONE")
        
        # ofile = sys.stdout
        # if outputfile != '':
        #     ofile = open(outputfile, 'w')

        # for i, top_k_list in enumerate(explanations_list):
        #     ofile.write('User question ' + str(i+1) + ':\n')
        #     for k, list_by_con in enumerate(top_k_list):
        #         for j in range(5):
        #             e = list_by_con[j]
        #             ofile.write('------------------------\n')
        #             print_str = ''
        #             e_tuple = data['df'].loc[data['df']['index'] == e[2]]
        #             e_tuple_str = ','.join(e_tuple.to_string(header=False,index=False,index_names=False).split('  ')[1:])
        #             ofile.write('Top ' + str(j+1) + ' explanation:\n')
        #             ofile.write('Constraint ' + str(e[1]+1) + ': [' + ','.join(constraints[e[1]][0]) + ']' + '[' + ','.join(constraints[e[1]][1]) + ']')
        #             ofile.write('\n')
        #             #ofile.write('Score: ' + str(e[0]))
        #             ofile.write('Score: ' + str(-e[0]))
        #             ofile.write('\n')
        #             ofile.write('(' + e_tuple_str + ')')
        #             ofile.write('\n')
            
        #         ofile.write('------------------------\n')

        ecf = self.config
        query_result_table = ecf.DEFAULT_RESULT_TABLE
        pattern_table = ecf.DEFAULT_PATTERN_TABLE
        user_question_file = ecf.DEFAULT_QUESTION_PATH
        outputfile = ''
        aggregate_column = ecf.DEFAULT_AGGREGATE_COLUMN
        conn = ecf.conn
        cur = ecf.cur
        
        # print(opts)
        start = time.clock()
        logger.info("start explaining ...")
        global_patterns, schema, global_patterns_dict = load_patterns(cur, pattern_table, query_result_table)
        logger.debug("loaded patterns from database")
        # category_similarity = CategorySimilarityMatrix(ecf.EXAMPLE_SIMILARITY_MATRIX_PATH, schema)
        category_similarity = CategorySimilarityNaive(cur=cur, table_name=query_result_table)
        # category_similarity = CategoryNetworkEmbedding(EXAMPLE_NETWORK_EMBEDDING_PATH, data['df'])
        #num_dis_norm = normalize_numerical_distance(data['df'])
        num_dis_norm = normalize_numerical_distance(cur=cur, table_name=query_result_table)

        # pf = PatternFinder(engine.connect(), query_result_table, fit=True, theta_c=0.5, theta_l=0.25, lamb=DEFAULT_LAMBDA, dist_thre=0.9,  supp_l=10,supp_g=1)
        Q, global_patterns, global_patterns_dict = load_user_question_from_file(
            global_patterns, global_patterns_dict, user_question_file, 
            schema, conn, cur, pattern_table, query_result_table, None, category_similarity)
        logger.debug("loaded user question from file")

        end = time.clock()
        print('Loading time: ' + str(end-start) + 'seconds')

        logger.debug("start finding explanations ...")
        
        start = time.clock()
        #regression_package = 'scikit-learn'
        regression_package = 'statsmodels'
        
        explanations_list, local_patterns_list, score_computing_time_list = find_explanation_regression_based(
            Q, global_patterns, global_patterns_dict, category_similarity, num_dis_norm,
            aggregate_column, conn, cur, 
            pattern_table, query_result_table, ecf
        )
        expl_end = time.time()
            
        end = time.clock()
        print('Total querying time: ' + str(end-start) + 'seconds')
        logger.debug("finding explanations ... DONE")

        ofile = sys.stdout
        if outputfile != '':
            ofile = open(outputfile, 'w')

        for i, top_k_list in enumerate(explanations_list):
            ofile.write('User question {} in direction {}: {}\n'.format(
                str(i+1), 'high' if Q[i]['dir'] > 0 else 'low', str(Q[i]['target_tuple']))
            )

            # print(1300, len(top_k_list))
            for j, e in enumerate(top_k_list):
                ofile.write('------------------------\n')
                ofile.write('Top ' + str(j+1) + ' explanation:\n')
                ofile.write(e.to_string())
                ofile.write('------------------------\n')
            #     print_str = ''
            #     # e_tuple_str = ','.join(e_tuple.to_string(header=False,index=False,index_names=False).split('  ')[1:])
            #     # print(e)
            #     if isinstance(e, dict):
            #         continue
            #     e_tuple_str = ','.join(map(str, e[2]))
            #     ofile.write('Top ' + str(j+1) + ' explanation:\n')
            #     # ofile.write('Constraint ' + str(e[1]+1) + ': [' + ','.join(global_patterns[e[1]][0]) + ']' + '[' + ','.join(global_patterns[e[1]][1]) + ']')
            #     # print(827, e[1], local_patterns_list[i][e[1]][0])
            #     # print(828, local_patterns_list[i][e[1]][1])
            #     # print(829, local_patterns_list[i][e[1]][2])
            #     # ofile.write('Constraint ' + str(e[1]+1) + ': [' + ','.join(local_patterns_list[i][e[1]][0]) + ']' + 
            #     #     '[' + ','.join(list(map(str, local_patterns_list[i][e[1]][1]))) + ']' +
            #     #     '[' + ','.join(list(map(str, local_patterns_list[i][e[1]][2]))) + ']')
                
            #     if e[5] == 1:
            #         ofile.write('From local pattern' + ': [' + ','.join(e[3][0]) + ']' + 
            #             '[' + ','.join(list(map(str, e[3][1]))) + ']' +
            #             '[' + ','.join(list(map(str, e[3][2]))) + ']' +
            #             '[' + e[3][4] + ']' + 
            #             (('[' + str(e[3][6].split(',')[0][1:]) + ']') if e[3][4] == 'const' else ('[' + str(e[3][7]) + ']'))
            #         )
            #         ofile.write('\ndrill down to\n' + ': [' + ','.join(e[4][0]) + ']' + 
            #             '[' + ','.join(list(map(str, e[4][1]))) + ']' +
            #             '[' + ','.join(list(map(str, e[4][2]))) + ']' + 
            #             '[' + e[4][4] + ']' + 
            #             (('[' + str(e[4][6].split(',')[0][1:]) + ']') if e[4][4] == 'const' else ('[' + str(e[4][7]) + ']'))
            #         )
            #     else:
            #         ofile.write('Directly from local pattern ' + ': [' + ','.join(e[3][0]) + ']' + 
            #             '[' + ','.join(list(map(str, e[3][1]))) + ']' +
            #             '[' + ','.join(list(map(str, e[3][2]))) + ']' +
            #             '[' + e[3][4] + ']' + 
            #             (('[' + str(e[3][6].split(',')[0][1:]) + ']') if e[3][4] == 'const' else ('[' + str(e[3][7]) + ']'))
            #         )
            #     ofile.write('\n')
            #     ofile.write('Score: ' + str(e[0]))
            #     ofile.write('\n')
            #     ofile.write('Distance: ' + str(e[1][0]))
            #     ofile.write('\n')
            #     # ofile.write('Simialriry: ' + str(e[1][1]))
            #     # ofile.write('\n')
            #     ofile.write('Outlierness: ' + str(e[1][2]))
            #     ofile.write('\n')
            #     ofile.write('Denominator: ' + str(e[1][3]))
            #     ofile.write('\n')
            #     ofile.write('(' + e_tuple_str + ')')
            #     ofile.write('\n')
            # # else:
            #     #     ofile.write('------------------------\n')
            #     #     ofile.write('Explanation:\n')
            #     #     ofile.write(str(list_by_pat) + '\n')
            # ofile.write('------------------------\n\n')
        ofile.close()

        for g_key in ecf.MATERIALIZED_DICT:
            for fv_key in ecf.MATERIALIZED_DICT[g_key]:
                dv_query = '''DROP VIEW IF EXISTS MV_{};'''.format(str(ecf.MATERIALIZED_DICT[g_key][fv_key]))
                cur.execute(dv_query)
                conn.commit()
        ecf.MATERIALIZED_DICT = dict()
        ecf.MATERIALIZED_CNT = 0

        
def main(argv=[]):
    try:
        opts, args = getopt.getopt(argv,"h:p:q:u:o:a",["help","ptable=","qtable=","ufile=","ofile=","aggregate_column="])
    except getopt.GetoptError:
        print('explanation.py -p <pattern_table> -q <query_result_table>  -u <user_question_file>\
         -o <outputfile> -a <aggregate_column>')
        sys.exit(2)
    # user_input_config = dict()
    user_input_config = ExplConfig()
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            print('explanation.py -p <pattern_table> -q <query_result_table> -u <user_question_file> \
                -o <outputfile> -a <aggregate_column>')
            sys.exit(2)    
        elif opt in ("-p", "--ptable"):
            user_input_config['pattern_table'] = arg
        elif opt in ("-q", "--qtable"):
            user_input_config['query_result_table'] = arg
        elif opt in ("-u", "--ufile"):
            user_input_config['user_question_file'] = arg
        elif opt in ("-o", "--ofile"):
            user_input_config['outputfile'] = arg
        elif opt in ("-a", "--aggcolumn"):
            user_input_config['aggregate_column'] = arg

    

    eg = ExplanationGenerator(user_input_config)
    # eg.doExplain()
    eg.initialize()
    # elist = eg.do_explain_online({'name': 'Jiawei Han', 'venue': 'kdd', 'year': 2007, 'sum_pubcount': 1, 'lambda': 0.2, 'direction': 'low'})
    # elist = eg.do_explain_online({'name': 'Kirsten Bergmann', 'venue': 'iva', 'sum_pubcount': 6.0, 'direction': 'high', 'lambda': 0.2})

    # elist = eg.do_explain_online({'primary_type': 'BATTERY', 'community_area': '26', 'year': '2011', 'count': 16, 'lambda': 0.2, 'direction': 'low'})
    for e in elist:
        print(e.to_string())


if __name__ == "__main__":
    main(sys.argv[1:])





from enum import Enum, unique

# ********************************************************************************
# commandline option possible datatypes
@unique
class OptionType(Enum):
    """
    Type of the value of a configuration option
    """
    Int = 'int',
    Float = 'float',
    String = 'string',
    Boolean = 'boolean'

# ********************************************************************************
# a commandline option
class ConfigOpt:
    """
    A commandline configuration option.
    """
    def __init__(self, longopt, shortopt=None, desc=None, hasarg=False, value=None, otype=OptionType.String, cfgFieldName=None):
        self.longopt = longopt
        self.shortopt = shortopt
        self.hasarg = hasarg
        self.description = desc
        self.value = value
        self.otype=otype
        self.cfgFieldName = cfgFieldName

    def helpString(self):
        helpMessage=''
        if self.shortopt != None:
            helpMessage='-' + self.shortopt + ' ,'
        helpMessage+= '--' + self.longopt
        if self.hasarg == True:
            helpMessage+= " <arg>"
        if self.description != None:
            helpMessage = '{:30}'.format(helpMessage)
            helpMessage+= " - " + self.description
        return helpMessage

    def castValue(self):
        if self.hasarg == True and self.value is not None:
            if self.otype == OptionType.Int:
                self.value =  int(self.value)
            elif self.otype == OptionType.Float:
                self.value =  float(self.value)
            elif self.otype == OptionType.Boolean:
                self.value = { 'True' : True,
                               'True' : True,
                               't' : True,
                               '1' : True,
                               'False' : False,
                               'false' : False,
                               'f' : False,
                               '0' : False}[self.value]
                self.value = bool(self.value)
            return self

# ********************************************************************************
# add support for accessing a class like a dictionary.
class DictLike:
    """
    Class used like an interface that makes a configuration class (data object) behave like a dictionary.
    """
    # overwrite __getitem__ to allow dictory style access to options
    def __getitem__(self, key):
        if key not in self.__dict__:
            raise AttributeError("No such attribute: " + key)
        return self.__dict__[key]

    # overwrite __setitem__ to allow dictory style setting of options
    def __setitem__(self,key,value):
        if key not in self.__dict__:
            raise AttributeError("No such attribute: " + key)
        self.__dict__[key] = value

    # get access to list of field names
    def getValidKeys(self):
        return self.__dict__

    # check whether configuration is valid
    def validateConfiguration(self):
        return True



from enum import Enum, unique
from capexplain.cl.cfgoption import DictLike
import logging

# logger for this module
log = logging.getLogger(__name__)

################################################################################
# Cmd types
@unique
class CmdTypes(Enum):
    Mine = 1,
    Explain = 2,
    Stats = 3,
    Help = 4,
    GUI = 5

# ********************************************************************************
# Information about a command for capexplain
class Command:

    def __init__(self, cmd, cmdstr, helpMessage, execute, options=None):
        self.cmd = cmd
        self.cmdstr = cmdstr
        self.options=options
        self.helpMessage=helpMessage
        self.execute = execute

    def helpString(self):
        return '{:30}- {}'.format(self.cmdstr,self.helpMessage)

    def __str__(self):
        return self.__dict__.__str__()

# ********************************************************************************
# multiple indexes for the options for a command
class CmdOptions:

    def constructOptions(self):
        self.shortopts = ''
        self.longopts = []
        self.shopt_map = {}
        self.longopt_map = {}
        self.cmdConfig = {}
        for opt in self.optionlist:
            self.cmdConfig[opt.longopt] = opt # mapping from configuration option names to configuration objects
            if opt.shortopt != None:
                self.shortopt_map['-' + opt.shortopt] = opt # map short option to configuration object
                self.shortopts+=opt.shortopt
                if opt.hasarg:
                    self.shortopts+=':'
            self.longopt_map['--' + opt.longopt] = opt
            if opt.hasarg:
                self.longopts.append(opt.longopt + '=')
            else:
                self.longopts.append(opt.longopt)

    def __init__(self, optionlist):
        self.optionlist = optionlist
        self.shortopt=''
        self.longopts=''
        self.longopt_map = {}
        self.shortopt_map = {}
        self.constructOptions()

    def setupConfig(self, config : DictLike):
        o = self
        for opt in o.cmdConfig:
            option = o.cmdConfig[opt]
            if option.value != None:
                key =  opt if (option.cfgFieldName is None) else option.cfgFieldName
                val = option.value
                log.debug("option: {}:{}".format(key,val))
                if key in config.getValidKeys():
                    config[key] = val
                else:
                    log.warning("unhandled config option <{}>".format(option.longopt))

    def setupConfigAndConnection(self, conn, config: DictLike):
        o = self
        for opt in o.cmdConfig:
            option = o.cmdConfig[opt]
            if option.value != None:
                key =  opt if (option.cfgFieldName is None) else option.cfgFieldName
                val = option.value
                log.debug("option: {}:{}".format(key,val))
                if key in conn.getValidKeys():
                    conn[key] = val
                elif key in config.getValidKeys():
                    config[key] = val
                else:
                    log.warning("unhandled config option <{}>".format(option.longopt))
        



import math
import geopy.distance


class CategorySimilarityNaive(object):
    """ Similarity measure for categorical attributes defined manually
    """
    def compute_similarity(self, col, val1, val2, aggr_col):
        # print(col, val1, val2)
        if col not in self.cate_cols:
            return 0
        if col in self.vector_dict:
            if val1 not in self.vector_dict[col] or val2 not in self.vector_dict[col]:
                return 0
            if col == 'community_area':
                dist_km = geopy.distance.vincenty(self.vector_dict[col][val1], self.vector_dict[col][val2]).km
                # print(val1, val2, 1.0 /
                    # math.exp(math.pow(dist_km,0.75)))
                sim = 7.0 / math.exp(math.pow(dist_km,0.75))
                if sim > 1:
                    sim = 1.0
                return sim
            dist = math.sqrt(sum(map(
                lambda x:(x[0]-x[1])*(x[0]-x[1]),
                zip(self.vector_dict[col][val1], self.vector_dict[col][val2]
            ))))
            return 1.0 / (1.0+dist)

        if val1 == val2:
            return 1.0
        else:
            return 0.0

    def __init__(self, cur, table_name, embedding_table_list=[]):
        type_query = '''SELECT column_name, data_type
            FROM information_schema.columns
            WHERE table_name = '{}';'''.format(table_name)
        cur.execute(type_query)
        res = cur.fetchall()
        print('Similarity Naive: ', res)
        self.cate_cols = {}
        self.vector_dict = {}
        for (col, dt) in res:
            if (dt == 'boolean' or dt.find('character') != -1) and (col != 'year' or len(embedding_table_list) == 0):
                self.cate_cols[col] = True

        if table_name.startswith('synthetic'):
            self.cate_cols['beat']  = True
            self.cate_cols['primary_type']  = True
            self.cate_cols['community_area']  = True
            self.cate_cols['district']  = True
            self.cate_cols['description']  = True
            self.cate_cols['location_description']  = True
            self.cate_cols['ward']  = True

        for (col, embedding_table_name) in embedding_table_list:
            self.vector_dict[col] = {}
            read_query = '''SELECT *  FROM {} ;'''.format(embedding_table_name)
            cur.execute(read_query)
            res = cur.fetchall()
            for (x, lat, log) in res:
                self.vector_dict[col][x] = (lat, log)        

    def is_categorical(self, col):
        return col in self.cate_cols


import datetime
import traceback
from itertools import combinations, permutations
from math import factorial
from tqdm import tqdm

def printException(ex,finfo):
    print(exceptionToString(ex,finfo))

def exceptionToString(ex,finfo):
    trace=traceback.format_stack()
    trace=trace[:-3]
    formatStr = '{}\n\nEXCEPTION:{}\n\n{}\n\nSTACK TRACE\n-----------\n{}'
    return formatStr.format(formatCurFileAndLine(finfo),
                            type(ex),
                            ex,
                            "\n".join(trace))

def formatCurFileAndLine(finfo):
    return 'FILE: {0.filename} LINE: {0.lineno}:'.format(finfo) 

def projection(t, cols):
    res = list()
    for c in cols:
        if c in t:
            res.append(t[c])
    return res
    # return list(map(lambda x: t[x], cols))

def get_F_value(F, t):
    return projection(t, F)

def get_V_value(V, t):
    return projection(t, V)

def is_float(s):
    try:
        if isinstance(s, float) or isinstance(s, int) or isinstance(s, str):
            float(s)
            return True
        else:
            return False
    except ValueError:
        return False

def is_integer(s):
    try:
        if isinstance(s, str) and s[-1] == 'L':
            int(s[0:-1])
        else:
            int(s)
        return True
    except ValueError:
        return False

def float_or_integer(s):
    if is_float(s):
        return float(s)
    elif is_integer(s):
        return int(s[0:-1]) if s[-1] == 'L' else int(s)
    else:
        return s

def tuple_column_to_str_in_where_clause(col_value):
    if is_float(col_value):
        return '=' + str(col_value)
    else:
        return "like '%" + col_value + "%'"

def normalize_numerical_distance(df=None, cur=None, table_name=''):
    '''
        get the min value, max value, and the range of each column in the data frame
    '''
    if df is not None:
        res = {}
        max_vals = df.max()
        min_vals = df.min()
        for col in df:
            if col == 'index':
                continue
            if df[col].dtype.kind != 'S' and df[col].dtype.kind != 'O':
                res[col] = {'max':{}, 'min':{}, 'range':{}}

                res[col]['max'] = float(max_vals[col])
                res[col]['min'] = float(min_vals[col])
                res[col]['range'] = res[col]['max'] - res[col]['min']
        return res
    else:
        res = {}
        column_name_query = "SELECT column_name FROM information_schema.columns where table_name='{}';".format(table_name)
        cur.execute(column_name_query)
        column_name = cur.fetchall()

        max_clause = ', '.join(map(lambda x: 'MAX(' + x + ')',
            map(lambda x: x[0] if x[0] != 'arrest' and x[0] != 'domestic' else '1', column_name)))
        max_query = "SELECT {} FROM {};".format(max_clause, table_name)
        cur.execute(max_query)
        max_vals = cur.fetchall()[0]
        min_clause = ', '.join(map(lambda x: 'MIN(' + x + ')',
            map(lambda x: x[0] if x[0] != 'arrest' and x[0] != 'domestic' else '1', column_name)))
        min_query = "SELECT {} FROM {};".format(min_clause, table_name)
        cur.execute(min_query)
        min_vals = cur.fetchall()[0]
        for idx, col_res in enumerate(column_name):
            col = col_res[0]
            res[col] = {'max':None, 'min':None, 'range':None}
            print('IN NORM', col, max_vals[idx], min_vals[idx])
            if is_float(max_vals[idx]) and is_float(min_vals[idx]):
                res[col]['max'] = float(max_vals[idx])
                res[col]['min'] = float(min_vals[idx])
                res[col]['range'] = res[col]['max'] - res[col]['min']
            else:
                if isinstance(max_vals[idx], datetime.date):
                    res[col]['max'] = max_vals[idx]
                    res[col]['min'] = min_vals[idx]
                    res[col]['range'] = res[col]['max'] - res[col]['min']
        return res

################################################################################
class CombinationsWithLen():
    """
    Wrapping a combinations generator to provide its length without having to iterate through all elements
    """

    def __init__(self,it,r):
        self._inLen = len(it)
        self._it = it
        self._r = r
        self._comb=combinations(it,r)

    def __iter__(self):
        for el in self._comb:
            yield el

    def __len__(self):
        return factorial(self._inLen) // (factorial(self._inLen - self._r)) // factorial(self._r)

def progress_iter(iter, showProgress=True, desc=None):
    if (showProgress == True):
        return tqdm(iter, desc=desc)
    else:
        return iter
    
################################################################################
class PermutationsWithLen():
    """
    Wrapping a permutations generator to provide its length without having to iterate through all elements
    """

    def __init__(self,it,r):
        self._inLen = len(it)
        self._it = it
        self._r = r
        self._perms=permutations(it,r)

    def __iter__(self):
        for el in self._perms:
            yield el

    def __len__(self):
        return factorial(self._inLen) // (factorial(self._inLen - self._r))



import sys
import pprint
import logging
import pandas as pd
from itertools import combinations, permutations
from time import sleep
import statsmodels.formula.api as sm
from psycopg2.extras import Json
from sklearn.linear_model import LinearRegression
from scipy.stats import chisquare, mode
from numpy import percentile, mean
from inspect import currentframe, getframeinfo
from tqdm import tqdm
from capexplain.utils import printException, CombinationsWithLen, progress_iter, PermutationsWithLen
from capexplain.pattern_miner.permtest import *
from capexplain.fd.fd import closure
from capexplain.cl.cfgoption import DictLike
from capexplain.cl.instrumentation import ExecStats


# setup logging
log = logging.getLogger(__name__)

class Explanation(object):
    expl_type = -1
    user_question_direction = 0
    tuple_value = dict()
    top_k_expl = -1
    relevent_pattern = None
    refinement_pattern = None
    score = -1
    distance = -1
    deviation = -1
    denominator = -1

    def __lt__(self, o):
        return self.score < o.score

    def __eq__(self, o):
        return self.score == o.score

    def __init__(self, et, score, distance, deviation, denominator, uqd, tv, top_k_expl, relevent_pattern, refinement_pattern):
        self.expl_type = et
        self.score = score
        self.distance = distance
        self.deviation = deviation
        self.denominator = denominator
        self.user_question_direction = uqd
        self.tuple_value = tv
        self.top_k_expl = top_k_expl
        self.relevent_pattern = relevent_pattern
        self.refinement_pattern = refinement_pattern
        # log.debug("created explanation:\n")

    def ordered_tuple_string(self):
        value_str = ''
        agg = ''
        for key in sorted(self.tuple_value.keys()):
            if key.startswith('count') or key.startswith('sum'):
                agg = key
            else:
                value_str += str(self.tuple_value[key]) + ','
        if agg in self.tuple_value:
            value_str += str(self.tuple_value[agg])
        return value_str

    def to_string(self):
        e_tuple_str = ','.join(map(str, self.tuple_value.values()))
        # res_str = 'Top ' + str(j+1) + ' explanation:\n'
        res_str = ''
        if self.expl_type == 1:
            res_str += 'From local pattern' + ': [' + ','.join(self.relevent_pattern[0]) + ']' +\
                '[' + ','.join(list(map(str, self.relevent_pattern[1]))) + ']' + \
                '[' + ','.join(list(map(str, self.relevent_pattern[2]))) + ']' +\
                '[' + self.relevent_pattern[4] + ']' + \
                (('[' + str(self.relevent_pattern[6].split(',')[0][1:]) + ']') if self.relevent_pattern[4] == 'const' else ('[' + str(self.relevent_pattern[7]) + ']'))
            res_str += '\ndrill down to\n' + ': [' + ','.join(self.refinement_pattern[0]) + ']' + \
                '[' + ','.join(list(map(str, self.refinement_pattern[1]))) + ']' + \
                '[' + ','.join(list(map(str, self.refinement_pattern[2]))) + ']' + \
                '[' + self.refinement_pattern[4] + ']' + \
                (('[' + str(self.refinement_pattern[6].split(',')[0][1:]) + ']') if self.refinement_pattern[4] == 'const' else ('[' + str(self.refinement_pattern[7]) + ']'))
        else:
            res_str += 'Directly from local pattern ' + ': [' + ','.join(self.relevent_pattern[0]) + ']' + \
                '[' + ','.join(list(map(str, self.relevent_pattern[1]))) + ']' + \
                '[' + ','.join(list(map(str, self.relevent_pattern[2]))) + ']' + \
                '[' + self.relevent_pattern[4] + ']' +  \
                (('[' + str(self.relevent_pattern[6].split(',')[0][1:]) + ']') if self.relevent_pattern[4] == 'const' else ('[' + str(self.relevent_pattern[7]) + ']'))
        res_str += '\n'
        res_str += 'Score: ' + str(self.score)
        res_str += '\n'
        res_str += 'Distance: ' + str(self.distance)
        res_str += '\n'
        # res_str += 'Simialriry: ' + str(e[1][1]))
        # res_str += '\n')
        res_str += 'Outlierness: ' + str(self.deviation)
        res_str += '\n'
        res_str += 'Denominator: ' + str(self.denominator)
        res_str += '\n'
        res_str += '(' + e_tuple_str + ')'
        res_str += '\n'
        
        return res_str





import sys
import pprint
import logging
import pandas as pd
from itertools import combinations
from time import sleep,time
import statsmodels.formula.api as sm
from psycopg2.extras import Json
from sklearn.linear_model import LinearRegression
from scipy.stats import chisquare, mode
from numpy import percentile, mean
from capexplain.pattern_miner.permtest import findpre
from inspect import currentframe, getframeinfo
from capexplain.utils import printException, CombinationsWithLen, progress_iter, PermutationsWithLen
from capexplain.fd.fd import closure
from capexplain.cl.cfgoption import DictLike
from capexplain.cl.instrumentation import ExecStats


# setup logging
log = logging.getLogger(__name__)

# ********************************************************************************
class MinerConfig(DictLike):
    """
    MinerConfig - configuration for the pattern mining algorithm
    """

    ALGORITHMS={'naive','naive_alternative','optimized'}
    STATS_MODELS={'statsmodels','sklearn'}
    
    def __init__(self,
                 conn=None,
                 table=None,
                 pattern_schema='pattern',
                 fit=True,
                 theta_c=0.1,
                 theta_l=0.1,
                 lamb=0.1,
                 dist_thre=0.99,
                 reg_package='statsmodels',
                 supp_l=10,
                 supp_g=100,
                 fd_check=True,
                 supp_inf=True,
                 manual_num=False,
                 algorithm='optimized',
                 showProgress=True):
        self.conn=conn
        self.theta_c=theta_c
        self.theta_l=theta_l
        self.lamb=lamb
        self.fit=fit
        self.dist_thre=dist_thre
        self.reg_package=reg_package
        self.supp_l=supp_l
        self.supp_g=supp_g
        self.fd_check=fd_check
        self.supp_inf=supp_inf
        self.manual_num=manual_num
        self.algorithm=algorithm
        self.table=table
        self.pattern_schema=pattern_schema
        self.showProgress=showProgress
        log.debug("created miner configuration:\n%s", self.__dict__)
    
    def validateConfiguration(self):
        """
        Validate configuration.
        """
        log.debug("validate miner configuration ...")
        if self.reg_package not in self.STATS_MODELS:
            log.warning('Invalid input for reg_package, reset to default')
            self.reg_package='statsmodels'
        if self.algorithm not in self.ALGORITHMS:
            log.warning('Invalid input for algorithm, reset to default')
            self.algorithm='optimized'
        if self.table is None:
            log.error("user did not specify table for mining")
            raise Exception('please specify a table to mine')
        log.debug("validation of miner configuration successful:\n%s", self.__dict__)
        return True

    def __str__(self):
        return self.__dict__.__str__()
    
    def printConfig(self):
        pprint.pprint(self.__dict__)


# ********************************************************************************
class MinerStats(ExecStats):
    """
    Statistics gathered during mining
    """

    TIMERS={'aggregate',
            'df',
            'regression',
            'insertion',
            'drop',
            'loop',
            'innerloop',
            'fd_detect',
            'check_fd',
            'total',
            'query_cube',
            'query_materializecube',
    }

    COUNTERS={'patcand.local',
              'patcand.global',
              'patterns.local',
              'patterns.global',
              'query.agg',
              'query.sort',
              'G',
              'F,V'
    }


# ********************************************************************************
class PatternFinder:
    """
    Mining patterns for an input relation. Patterns are stored in a table X
    """

    config=None
    stats=None

    glob=None #global patterns
    num_rows=None #number of rows of config.table
    num=None #numeric attributes
    summable=None#summable attributes
    n=None#number of attributes
    grouping_attr=None #attributes can be in group
    schema=None #list of all attributes
    time=None #record running time for each section
    superkey=None #used to track compound key
    failedf=None #used to apply support inference
    group_rows=None #track # of rows for each group, used for fd check
    
    def __init__(self, config : MinerConfig):
        self.config = config
        self.initDataStructures()
        log.debug("initialized datastructures")
        self.fetchInputTableInfo()
        log.info("fetched basic table information - ready to mine")

    def fetchInputTableInfo(self):
        try:
            self.schema=list(pd.read_sql("SELECT * FROM "+self.config.table+" LIMIT 1",self.config.conn))
        except Exception as ex:
            printException(ex, getframeinfo(currentframe()))
            sys.exit(2)
        self.n=len(self.schema)
        log.debug("fetched schema for input table: %s with %d attributes", self.schema, self.n)        
            
        self.num_rows=pd.read_sql("SELECT count(*) as num from "+self.config.table,self.config.conn)['num'][0]
        log.debug("input table has %d rows", self.num_rows)

        #check uniqueness, grouping_attr contains only non-unique attributes
        unique=pd.read_sql("SELECT "+','.join(["count(distinct "+attr+") as "+attr for attr in self.schema])+ " FROM "+self.config.table,
                                              self.config.conn)                            
        for attr in self.schema:
            #aggresive approach:
            n_distinct=unique[attr][0]
            if n_distinct>=self.num_rows*self.config.dist_thre:
                continue
            self.grouping_attr.append(attr)
            self.group_rows[frozenset([attr])]=n_distinct

        log.debug("possible grouping attributes: %s", self.grouping_attr)
                
        for col in self.schema:
            try: # Boris: this may fail, better to get the datatype from the catalog and have a list of numeric datatypes to check for
                self.config.conn.execute("SELECT CAST("+col+" AS NUMERIC) FROM "+self.config.table)
                self.num.append(col)
            except:
                continue
        if self.config.manual_num:
            self.setNumeric()
        else:
            self.summable=self.num
        log.debug("tables has numerical attributes %s and summable attributes", self.num, self.summable)

    def progress(self, iter, desc=None):
        return progress_iter(iter=iter, desc=desc, showProgress=self.config.showProgress)
        
    def initDataStructures(self):
        self.stats = MinerStats()
        self.superkey=set()        
        self.num=[]
        self.grouping_attr=[]
        self.group_rows={}
        
    def setNumeric(self):
        print("Current Numeric: ")
        print(self.num)
        if input('Redefine? (y/n, default n)')=='y':
            num=[]
            for attr in self.num:
                keep=input('Keep '+attr+' as Numeric?(y/n,default y)')
                if keep=='n':
                    continue
                else:
                    num.append(attr)
            self.num=num
            self.summable=[attr for attr in self.summable if attr in num]
        print("Current Summable: ")
        print(self.summable)
        if input('Redefine? (y/n, default n)')=='y':
            summable=[]
            for attr in self.summable:
                keep=input('Keep '+attr+' as Summable?(y/n,default y)')
                if keep=='n':
                    continue
                else:
                    summable.append(attr)
            self.summable=summable
        
    def validateFd(self,group,division=None):
        '''
        check if we want to ignore f=group[:division], v=group[division:]
        type group: tuple of strings
        
        return True if group is valid
        return False if we decide to ignore it
        
        if division=None, check all possible divisions, and return boolean[]
        '''
        if division:
            if not self.config.fd_check:
                return True
            if division==1:
                return True
            row=self.group_rows[frozenset(group[:division])]
            for i in group[:division]:
                if self.group_rows[frozenset([k for k in group[:division] if k!=i])]>=row*self.config.dist_thre:
                    return False
            return True
        else:
            n=len(group)
            ret=[self.validateFd(group,i) for i in range(1,n)] #division from 1 to n-1
            return ret
            
    def findPattern(self,user=None):
        if (self.config.showProgress):
            print(" MINING PATTERNS FOR: {}".format(self.config.table))
#       self.pc=PC.PatternCollection(list(self.schema))
        self.glob=[]#reset self.glob
        self.createTable(self.config.pattern_schema)
        self.stats.startTimer('total')
        if not user:
            grouping_attr=self.grouping_attr
            aList=self.summable+['*']
        else:
            log.debug("user provided list of group-by and aggregation input attributes %s", user)
            grouping_attr=user['group']
            aList=user['a']
        log.debug("mine patterns for group-by attrs %s and aggregate input attributes %s", grouping_attr, aList)
        
        if self.config.algorithm=='naive':# CUBE ALGORITHM
            for a in self.progress(aList, desc='agg functions'):
                if not user:
                    if a=='*':#For count, only do a count(*)
                        agg="count"
                    else: #a in self.num
                        agg="sum"
                else:
                    agg=user['agg']
    
                log.debug("consider aggregation %s(%s) ", agg, a)
                #for agg in aggList :
                cols=[col for col in grouping_attr if col!=a]
                n=len(cols)
                self.formCube(a, agg, cols)
                for size in self.progress(range(min(4,n),1,-1), desc='pattern size'):
                    combs=CombinationsWithLen(cols,size)
                    for group in self.progress(combs, desc='F,V combinations'):#comb=f+v
                        self.stats.incr('G')
                        log.debug("consider group-by attributes %s", group)
                        for fsize in self.progress(range(1,len(group)), desc='F size'):
                            fs=CombinationsWithLen(group,fsize)
                            for f in self.progress(fs, desc='F'):
                                self.stats.incr('F,V')
                                self.stats.incr('patcand.global')
                                log.debug("consider group-by attributes %s with F=%s", group, f)
                                self.fit_naive(f,group,a,agg,cols)
                self.dropCube()
        else:#self.config.algorithm=='optimized' or self.config.algorithm=='naive_alternative'
            self.failedf=set()
            for size in self.progress(range(2,min(4,len(grouping_attr))+1),desc='pattern size'):
                for group in self.progress(combinations(grouping_attr,size), desc='group'):
                    for i in self.superkey:
                        if set(i).issubset(group):
                            break
                    else:#doesn't contain superkey
                        aggList=[]
                        for a in aList:
                            if a in group:
                                continue
                            if a=='*':
                                aggList.append('count')
                            else:
                                aggList.append('sum_'+a+'')
                        if len(aggList)==0:
                            continue
                        self.aggQuery(group, aList)
                        self.stats.startTimer('fd_detect')
                        isSuperkey=False
                        if self.config.fd_check:
                            cur_rows=pd.read_sql('SELECT count(*) as num FROM agg',con=self.config.conn)['num'][0]
                            if cur_rows>=self.num_rows*self.config.dist_thre:
                                self.superkey.add(group)
                                isSuperkey=True
                            self.group_rows[frozenset(group)]=cur_rows
                        self.stats.stopTimer('fd_detect')
                        if not isSuperkey:
                            if self.config.algorithm=='naive_alternative':
                                for fsize in self.progress(range(size-1,0,-1),desc='Fsize'):
                                    for f in self.progress(combinations(group,fsize),desc='F'):
                                        self.stats.startTimer('df')
                                        df=pd.read_sql('SELECT * FROM agg ORDER BY '+','.join(f),con=self.config.conn)
                                        self.stats.stopTimer('df')
                                        grouping=tuple([fattr for fattr in f]+[vattr for vattr in group if vattr not in f])
                                        division=len(f)
                                        self.fitmodel(df, grouping, aggList, division)  
                            else:#algorithm='optimized'
                                for i in self.progress(range(len(group)),desc='Group size'):
                                    perm=group[i:]+group[:i]
                                    self.stats.startTimer('df')
                                    df=pd.read_sql('SELECT * FROM agg ORDER BY '+','.join(perm[:-1]),con=self.config.conn)
                                    self.stats.stopTimer('df')
                                    self.fitmodel(df, perm, aggList, None)
                                if len(group)==4:
                                    for f in [(group[0],group[2]),(group[1],group[3])]:
                                        grouping=tuple([fattr for fattr in f]+[vattr for vattr in group if vattr not in f])
                                        division=len(f)
                                        self.stats.startTimer('check_fd')
                                        fd_val=self.validateFd(grouping, division)
                                        self.stats.stopTimer('check_fd')
                                        if fd_val:
                                            self.stats.startTimer('df')
                                            df=pd.read_sql('SELECT * FROM agg ORDER BY '+','.join(f),con=self.config.conn)
                                            self.stats.stopTimer('df')
                                            self.fitmodel(df, grouping, aggList, division)       
                        self.dropAgg()    
        if self.glob:
            self.stats.startTimer('insertion')
            self.config.conn.execute("INSERT INTO "+self.config.pattern_schema+"."+self.config.table+"_global values"+','.join(self.glob))
            self.stats.stopTimer('insertion')
        self.stats.stopTimer('total')
        self.insertTime('\''+self.config.table+'_num_global:'+str(len(self.glob))+' algo:'+self.config.algorithm+'\'')
        log.warning("pattern mining finished: time stats:\n\n%s", self.stats.formatStats())
        if (self.config.showProgress):
            sleep(0.5)
            print("\n\n\n\n\nFINISHED MINING PATTERNS FOR {} found {} global patterns based on {} local patterns\n\n"
                  .format(self.config.table, self.stats.getCounter('patterns.global'), self.stats.getCounter('patterns.local')))
        
    def formCube(self, a, agg, attr):
        self.stats.startTimer('query_materializecube')
        self.stats.incr('query.agg')
        group=",".join(["CAST("+num+" AS NUMERIC)" for num in attr if num in self.num]+
                        [cat for cat in attr if cat not in self.num])
        grouping=",".join(["CAST("+num+" AS NUMERIC), GROUPING(CAST("+num+" AS NUMERIC)) as g_"+num
                        for num in attr if num in self.num]+
            [cat+", GROUPING("+cat+") as g_"+cat for cat in attr if cat not in self.num])
        if a in self.num:
            qa="CAST("+a+" AS NUMERIC)"
        else:
            qa=a
        if agg=='count':
            name='count'
        else:
            name='sum_'+a
        self.config.conn.execute("DROP TABLE IF EXISTS cube")
        query="CREATE TABLE cube AS SELECT "+agg+"("+qa+") AS \""+name+"\", "+group+','+grouping+" FROM "+self.config.table+"\
        "+"GROUP BY CUBE("+group+") having "+'+'.join(['grouping('+att+')' if att not in self.num 
                                else "GROUPING(CAST("+att+" AS NUMERIC))" for att in attr])+'>='+str(len(attr)-4)
        log.debug("Materialize CUBE:\n\n%s", query)
        self.config.conn.execute(query)        
        self.stats.stopTimer('query_materializecube')

    def dropCube(self):
        log.debug("DROP materialized CUBE")
        self.config.conn.execute("DROP TABLE cube;")
        
    def cubeQuery(self, g, f, cols):
        self.stats.startTimer('query_cube')
        self.stats.incr('query.sort')
        res=" and ".join(["g_"+a+"=0" for a in g])
        if len(g)<len(cols):
            unused=" and ".join(["g_"+b+"=1" for b in cols if b not in g])
            res=res+" and "+unused
        query= "SELECT * FROM cube where "+res+" ORDER BY "+",".join(f)
        log.debug("Run query over materialized CUBE:\n\n %s", query)
        self.stats.stopTimer('query_cube')
        return query
    
    def fit_naive(self,f,group,a,agg,cols):
        self.failedf=set()#to not trigger error
        fd=pd.read_sql(self.cubeQuery(group, f, cols),self.config.conn)
        g=tuple([att for att in f]+[attr for attr in group if attr not in f])
        division=len(f)
        if a=='*':
            aggList=['count']
        else:
            aggList=[agg+'_'+a]
        self.fitmodel_with_division(fd, g, aggList, division)
        
    def findPattern_inline(self,group,a,agg):
        #loop through permutations of group
        user={'group':group,'a':[a],'agg':agg}
        self.findPattern(user)
        
    def aggQuery(self, g, aList):
        self.stats.startTimer('aggregate')
        group=",".join(["CAST("+att+" AS NUMERIC)" if att in self.num else att for att in g])
        agg=[]
        for a in aList:
            if a in g:
                continue
            if a=='*':
                agg.append('count(*) AS \"count\"')
            else:
                agg.append('sum(CAST('+a+' AS NUMERIC)) AS \"sum_'+a+'\"')
        query="CREATE TEMP TABLE agg as SELECT "+group+","+','.join(agg)+" FROM "+self.config.table+" GROUP BY "+group
        self.config.conn.execute(query)
        self.stats.stopTimer('aggregate')
    
    def rollupQuery(self, group, d_index, agg):
        self.stats.startTimer('aggregate')
        grouping=",".join([attr+", GROUPING("+attr+") as g_"+attr for attr in group[:d_index]])
#        gsets=','.join(['('+','.join(group[:prefix])+')' for prefix in range(d_index,pre,-1)])
        self.config.conn.execute('CREATE TEMP TABLE grouping AS '+
                        'SELECT '+grouping+', SUM('+agg+') as '+agg+
#                        ' FROM agg GROUP BY GROUPING SETS('+gsets+')'+
                        ' FROM agg GROUP BY ROLLUP('+','.join(group)+')'+
                        ' ORDER BY '+','.join(group[:d_index]))
        self.stats.stopTimer('aggregate')
        
    def dropRollup(self):
        self.stats.startTimer('drop')
        self.config.conn.execute('DROP TABLE grouping')
        self.stats.stopTimer('drop')
        
    def dropAgg(self):
        self.stats.startTimer('drop')
        self.config.conn.execute('DROP TABLE agg')
        self.stats.stopTimer('drop')
        
    def fitmodel(self, fd, group, aggList, division):
        self.stats.startTimer('loop')
        if division:
            self.fitmodel_with_division(fd, group, aggList, division)
        else:
            self.fitmodel_no_division(fd, group, aggList)
        self.stats.stopTimer('loop')
            
    def fitmodel_no_division(self, fd, group, aggList):
        size=len(group)-1
        oldKey=None
        oldIndex=[0]*size
        num_f=[0]*size
        valid_l_f=[0]*size
        valid_c_f=[0]*size
        f=[list(group[:i]) for i in range(1,size+1)]
        v=[list(group[j:]) for j in range(1,size+1)]
        supp_valid=[group[:i] not in self.failedf for i in range(1,size+1)]
        f_dict=[{} for i in range(1,size+1)]
        self.stats.startTimer('check_fd')
        fd_valid=self.validateFd(group)
        self.stats.startTimer('check_fd')
        sample_invalid=[{} for i in range(size)]
        global_dev_pos=[{} for i in range(size)]
        global_dev_neg=[{} for i in range(size)]
        for i in range(size):
            for agg in aggList:
                global_dev_pos[i][agg]={}
                global_dev_pos[i][agg]['l']=float('-inf')
                global_dev_pos[i][agg]['c']=float('-inf')
                global_dev_neg[i][agg]={}
                global_dev_neg[i][agg]['l']=float('inf')
                global_dev_neg[i][agg]['c']=float('inf')
        if not any(fd_valid) or not any(supp_valid):
            return
        pattern=[]
        def fit(df,fval,i,n):
            if not self.config.fit:
                return
            self.stats.startTimer('regression')
            for agg in aggList:
                nonlocal global_dev_pos,global_dev_neg
                if agg not in sample_invalid[i] or 'c' not in sample_invalid[i][agg]:
                    avg=mean(df[agg])
                    describe=[avg,mode(df[agg]),percentile(df[agg],25)
                              ,percentile(df[agg],50),percentile(df[agg],75)]
                    dev_pos=max(df[agg])-avg
                    dev_neg=min(df[agg])-avg
                    #fitting constant
                    theta_c=chisquare(df[agg].dropna())[1]
                    if theta_c>self.config.theta_c:
                        nonlocal valid_c_f
                        try:
                            valid_c_f[i][agg]+=1
                        except KeyError:
                            valid_c_f[i][agg]=1
                        global_dev_pos[i][agg]['c']=max(global_dev_pos[i][agg]['c'],dev_pos)
                        global_dev_neg[i][agg]['c']=min(global_dev_neg[i][agg]['c'],dev_neg)
                        #self.pc.add_local(f,oldKey,v,a,agg,'const',theta_c)
                        pattern.append(self.addLocal(f[i],fval,v[i],agg,'const',theta_c,describe,'NULL',
                                                     dev_pos,dev_neg))
                        self.stats.incr('patterns.local')
              
                #fitting linear
                if agg not in sample_invalid[i] or 'l' not in sample_invalid[i][agg]:
                    if  theta_c!=1 and ((self.config.reg_package=='sklearn' and all(attr in self.num for attr in v[i])
                                        or
                                        (self.config.reg_package=='statsmodels' and all(attr in self.num for attr in v[i])))):
        
                        if self.config.reg_package=='sklearn':   
                            lr=LinearRegression()
                            lr.fit(df[v[i]],df[agg])
                            theta_l=lr.score(df[v[i]],df[agg])
                            theta_l=1-(1-theta_l)*(n-1)/(n-len(v[i])-1)
                            param=lr.coef_.tolist()
                            param.append(lr.intercept_.tolist())
                            param="'"+str(param)+"'"
                        else: #statsmodels
                            if n<=len(v[i])+1:#negative R^2 for sure
                                return
                            lr=sm.ols(agg+'~'+'+'.join([attr if attr in self.num else 'C('+attr+')' for attr in v[i]])
                                      ,data=df,missing='drop').fit()
                            theta_l=lr.rsquared_adj
                            param=Json(dict(lr.params))
                            dev_pos=max(lr.resid)
                            dev_neg=min(lr.resid)
                        
                        if theta_l and theta_l>self.config.theta_l:
                            nonlocal valid_l_f
                            try:
                                valid_l_f[i][agg]+=1
                            except KeyError:
                                valid_l_f[i][agg]=1
                            global_dev_pos[i][agg]['l']=max(global_dev_pos[i][agg]['l'],dev_pos)
                            global_dev_neg[i][agg]['l']=min(global_dev_neg[i][agg]['l'],dev_neg)
                            pattern.append(self.addLocal(f[i],fval,v[i],agg,'linear',theta_l,describe,param,
                                                         dev_pos,dev_neg))
                            self.stats.incr('patterns.local')
                        
            self.stats.stopTimer('regression')
            
        self.stats.startTimer('innerloop')
 
        for tup in fd.itertuples():
            
            position=None
            if oldKey:
                for i in range(size):
                    if getattr(tup,group[i])!=getattr(oldKey,group[i]):
                        position=i
                        break 
            
            if position is not None:
                index=tup.Index
                for i in range(position,size):
                    if not fd_valid[i] or not supp_valid[i]:
                        continue
                    n=index-oldIndex[i]
                    if n>=self.config.supp_l:
                        num_f[i]+=1
                        fval=tuple([getattr(oldKey,j) for j in f[i]])
                        f_dict[i][fval]=[oldIndex[i],index]
                    oldIndex[i]=index
                    
            oldKey=tup
            
        if oldKey:
            for i in range(size):
                if not fd_valid[i] or not supp_valid[i]:
                    continue
                n=oldKey.Index-oldIndex[i]+1
                if n>self.config.supp_l:
                    num_f[i]+=1
                    fval=tuple([getattr(oldKey,j) for j in f[i]])
                    f_dict[i][fval]=[oldIndex[i]]
        self.stats.stopTimer('innerloop')
        
        for i in range(size):
            if len(f_dict[i])<self.config.supp_g:
                if self.config.supp_inf:#if toggle is on
                    self.failedf.add(tuple(f[i]))
                supp_valid[i]=False
            else:
                for fval in f_dict[i]:
                    indices=f_dict[i][fval]
                    if len(indices)==2: #indices=[oldIndex,index]
                        fit(fd[indices[0]:indices[1]],fval,i,indices[1]-indices[0])
                    else: #indices=[oldIndex]
                        fit(fd[indices[0]:],fval,i,oldKey.Index-indices[0]+1)
                    
        #sifting global            
        for i in range(size):
            if not fd_valid[i] or not supp_valid[i]:
                    continue
            for agg in valid_c_f[i]:
                if agg in sample_invalid[i] and 'c' in sample_invalid[i][agg]:
                    continue
                lamb_c=valid_c_f[i][agg]/num_f[i]
                if valid_c_f[i][agg]>=self.config.supp_g and lamb_c>self.config.lamb:
                    self.glob.append(self.addGlobal(f[i],v[i],agg,'const',self.config.theta_c,lamb_c,
                                                    global_dev_pos[i][agg]['c'],global_dev_neg[i][agg]['c']))
            
            for agg in valid_l_f[i]:
                if agg in sample_invalid[i] and 'l' in sample_invalid[i][agg]:
                    continue
                lamb_l=valid_l_f[i][agg]/num_f[i]
                if valid_l_f[i][agg]>=self.config.supp_g and lamb_l>self.config.lamb:
                    self.glob.append(self.addGlobal(f[i],v[i],agg,'linear',self.config.theta_l,lamb_l,
                                                    global_dev_pos[i][agg]['l'],global_dev_neg[i][agg]['l']))
        
        if not self.config.fit:
            return 

        if pattern:
            self.stats.startTimer('insertion')
            self.config.conn.execute("INSERT INTO "+self.config.pattern_schema+"."+self.config.table+"_local values"+','.join(pattern))        
            self.stats.stopTimer('insertion')

    def fitmodel_with_division(self, fd, group, aggList, division):  
        if group[:division] in self.failedf:
            log.debug("do not consider already failed F=%s", group[:division])
            return
        
        oldKey=None
        oldIndex=0
        num_f=0
        valid_l_f={}
        valid_c_f={}
        f=list(group[:division])
        v=list(group[division:])
        log.debug("fitting patterns for F=%s, V=%s", f, v)
        sample_invalid={}
        global_dev_pos={}
        global_dev_neg={}
        for agg in aggList:
            global_dev_pos[agg]={}
            global_dev_pos[agg]['l']=float('-inf')
            global_dev_pos[agg]['c']=float('-inf')
            global_dev_neg[agg]={}
            global_dev_neg[agg]['l']=float('inf')
            global_dev_neg[agg]['c']=float('inf')
        #df:dataframe n:length    
        pattern=[]
        def fit(df,f,fval,v,n):
            self.stats.incr('patcand.local')
            if not self.config.fit:
                return
            log.debug("do regression for F=%s, f=%s", f, fval)
            self.stats.startTimer('regression')
            for agg in aggList:
                nonlocal global_dev_pos,global_dev_neg
                if agg not in sample_invalid or 'c' not in sample_invalid[agg]:
                    avg=mean(df[agg])
                    describe=[avg,mode(df[agg]),percentile(df[agg],25)
                                                  ,percentile(df[agg],50),percentile(df[agg],75)]
                    dev_pos=max(df[agg])-avg
                    dev_neg=min(df[agg])-avg                    
                    #fitting constant
                    theta_c=chisquare(df[agg].dropna())[1]
                    if theta_c>self.config.theta_c:
                        nonlocal valid_c_f
                        try:
                            valid_c_f[agg]+=1
                        except KeyError:
                            valid_c_f[agg]=1
                        global_dev_pos[agg]['c']=max(global_dev_pos[agg]['c'],dev_pos)
                        global_dev_neg[agg]['c']=min(global_dev_neg[agg]['c'],dev_neg)

                        pattern.append(self.addLocal(f,fval,v,agg,'const',theta_c,describe,'NULL',
                                                     dev_pos,dev_neg))
                        self.stats.incr('patterns.local')
                
                #fitting linear
                if agg not in sample_invalid or 'l' not in sample_invalid[agg]:
                    if  theta_c!=1 and ((self.config.reg_package=='sklearn' and all(attr in self.num for attr in v)
                                        or
                                        (self.config.reg_package=='statsmodels' and all(attr in self.num for attr in v)))):
        
                        if self.config.reg_package=='sklearn': 
                            lr=LinearRegression()
                            lr.fit(df[v],df[agg])
                            theta_l=lr.score(df[v],df[agg])
                            theta_l=1-(1-theta_l)*(n-1)/(n-len(v)-1)
                            param=lr.coef_.tolist()
                            param.append(lr.intercept_.tolist())
                            param="'"+str(param)+"'"
                        else: #statsmodels
                            if n<=len(v)+1:#negative R^2 for sure
                                return
                            lr=sm.ols(agg+'~'+'+'.join([attr if attr in self.num else 'C('+attr+')' for attr in v]),
                                      data=df,missing='drop').fit()
                            #theta_l=lr.rsquared_adj
                            theta_l=chisquare(df[agg],lr.predict())[1]
                            param=Json(dict(lr.params))
                            dev_pos=max(lr.resid)
                            dev_neg=min(lr.resid)
                            
                        if theta_l and theta_l>self.config.theta_l:
                            nonlocal valid_l_f
                            try:
                                valid_l_f[agg]+=1
                            except KeyError:
                                valid_l_f[agg]=1
                            global_dev_pos[agg]['l']=max(global_dev_pos[agg]['l'],dev_pos)
                            global_dev_neg[agg]['l']=min(global_dev_neg[agg]['l'],dev_neg)
                            pattern.append(self.addLocal(f,fval,v,agg,'linear',theta_l,describe,param,
                                                         dev_pos,dev_neg))
                            self.stats.incr('patterns.local')
                    
            self.stats.stopTimer('regression')
        
        self.stats.startTimer('innerloop')
        change=False
        f_dict={}
        for tup in fd.itertuples():
            if oldKey:
                change=any([getattr(tup,attr)!=getattr(oldKey,attr) for attr in f])
            if change:
                index=tup.Index
                n=index-oldIndex
                if n>=self.config.supp_l:
                    num_f+=1
                    fval=tuple([getattr(oldKey,j) for j in f])
                    log.debug("local support high enough for %s (%d > %d)", fval, n, self.config.supp_l)
                    f_dict[fval]=[oldIndex,index]
                oldIndex=index
            oldKey=tup
            
        if oldKey:
            n=oldKey.Index-oldIndex+1
            if n>=self.config.supp_l:
                num_f+=1
                #fit(fd[oldIndex:],f,v,n)
                fval=tuple([getattr(oldKey,j) for j in f])
                f_dict[fval]=[oldIndex]
        self.stats.stopTimer('innerloop')
        
        if len(f_dict)<self.config.supp_g:
            if self.config.supp_inf:#if toggle is on
                self.failedf.add(group[:division])
            log.debug("global support threshold not reached for pattern F=%s, V=%s (%d < %d)", f, v, len(f_dict), self.config.supp_g)
            return
        else:
            for fval in f_dict:
                indices=f_dict[fval]
                if len(indices)==2: #indices=[oldIndex,index]
                    fit(fd[indices[0]:indices[1]],f,fval,v,indices[1]-indices[0])
                else: #indices=[oldIndex]
                    fit(fd[indices[0]:],f,fval,v,oldKey.Index-indices[0]+1)
        
        if pattern:
            log.debug("insert local pattern %s", pattern)
            self.stats.startTimer('insertion')
            self.config.conn.execute("INSERT INTO "+self.config.pattern_schema+"."+self.config.table+"_local values"+','.join(pattern))
            self.stats.stopTimer('insertion')
        
        for agg in valid_c_f:
            if agg in sample_invalid and 'c' in sample_invalid[agg]:
                continue
            lamb_c=valid_c_f[agg]/num_f
            if valid_c_f[agg]>=self.config.supp_g and lamb_c>self.config.lamb:
                log.info("found global pattern P = (%s, %s, %s, const)", f, v, agg)
                self.glob.append(self.addGlobal(f,v,agg,'const',self.config.theta_c,lamb_c,
                                                global_dev_pos[agg]['c'],global_dev_neg[agg]['c']))
                self.stats.incr('patterns.global')
            else:
                log.info("global pattern candidate P = (%s, %s, %s, const) does not hold", f, v, agg)
             
        for agg in valid_l_f:
            if agg in sample_invalid and 'l' in sample_invalid[agg]:
                continue
            lamb_l=valid_l_f[agg]/num_f
            if valid_l_f[agg]>=self.config.supp_g and lamb_l>self.config.lamb:
                log.info("found global pattern P = (%s, %s, %s, linear)", f, v, agg)
                self.glob.append(self.addGlobal(f,v,agg,'linear',self.config.theta_l,lamb_l,
                                                global_dev_pos[agg]['l'],global_dev_neg[agg]['l']))
                self.stats.incr('patterns.global')            
            else:
                log.info("global pattern candidate P = (%s, %s, %s, linear) does not hold", f, v, agg)
         
                          
    def addLocal(self,f,f_val,v,agg,model,theta,describe,param,dev_pos,dev_neg):#left here
        f='ARRAY'+str(list(f)).replace('"','')
        f_val='ARRAY'+str([str(val).replace("'","") for val in f_val]).replace('"','')
        v='ARRAY'+str(list(v)).replace('"','')
        agg="'"+agg+"'"
        model="'"+model+"'"
        theta="'"+str(theta)+"'"
        describe="'"+str(describe).replace("'","")+"'"
        return '('+','.join([f,f_val,v,agg,model,theta,describe,str(param),str(dev_pos),str(dev_neg)])+')'
    
    
    def addGlobal(self,f,v,agg,model,theta,lamb,dev_pos,dev_neg):
        f='ARRAY'+str(list(f)).replace('"','')
        v='ARRAY'+str(list(v)).replace('"','')
        agg="'"+agg+"'"
        model="'"+model+"'"
        theta="'"+str(theta)+"'"
        lamb="'"+str(lamb)+"'"
        return '('+','.join([f,v,agg,model,theta,lamb,str(dev_pos),str(dev_neg)])+')'
         
    
    def createTable(self, pattern_schema):
        if self.config.reg_package=='sklearn':
            ptype='varchar'
        else:
            ptype='json'
        
        self.config.conn.execute('CREATE SCHEMA IF NOT EXISTS '+pattern_schema)
        
        self.config.conn.execute('DROP TABLE IF EXISTS '+pattern_schema+'.'+self.config.table+'_local;')    
        self.config.conn.execute('create table IF NOT EXISTS '+pattern_schema+'.'+self.config.table+'_local('+
                     'fixed varchar[],'+
                     'fixed_value varchar[],'+
                     'variable varchar[],'+
                     'agg varchar,'+
                     'model varchar,'+
                     'theta float,'+
                     'stats varchar,'+
                     'param '+ptype+','+
                     'dev_pos float,'+
                     'dev_neg float);')
        
        self.config.conn.execute('DROP TABLE IF EXISTS '+pattern_schema+'.'+self.config.table+'_global')
        self.config.conn.execute('create table IF NOT EXISTS '+pattern_schema+'.'+self.config.table+'_global('+
                     'fixed varchar[],'+
                     'variable varchar[],'+
                     'agg varchar,'+
                     'model varchar,'+
                     'theta float,'+
                     'lambda float, '+
                     'dev_pos float,'+
                     'dev_neg float);')
        
        attr=''
        for key in self.time:
            attr+=key+' varchar,'
        self.config.conn.execute('create table IF NOT EXISTS time_naive('+
                          'id serial primary key,'+
                          attr+
                          'description varchar);')
        log.debug("done creating tables")
        
        
    def insertTime(self,description):
        attributes=list(self.stats.time)
        values=[str(self.stats.time[i]) for i in attributes]
        attributes.append('description')
        values.append(description)
        self.config.conn.execute('INSERT INTO time_detail_fd('+','.join(attributes)+') values('+','.join(values)+')')




import textwrap



string = '''     
Explanation for why sum(publcount) is lower 
than expected for name=Ax,venue=ICDE,year=2009:
Even though like many other authors, year predicts 
sum(pubcount) for author = Ax (left graph), Ax's publication
in venue=icde, year=2009 is low which maybe explained by his higher than usual
number of publications in year = 2008, venue = icde(right graph)
    	'''

lists = textwrap.wrap(string)

formatted_string = '\n'.join(lists)

print(formatted_string)




def closure(fd,n,a):
    '''
    return the closure of a, attributes are represented as corresponding integer indices in all schema
    type fd: list of tuple where each tuple = (lhs,rhs) where lhs,rhs are lists of int
    type n: length of all attributes
    type a: tuple/list/set of int
    '''
    if not a:
        return set()
    m=len(fd)
    c=[0]*m
    lhs=[[] for i in range(n)]
    rhs=[[] for i in range(m)]
    aplus=set()
    todo=list(a)
    
    for i in range(m):
        left=fd[i][0]
        right=fd[i][1]
        c[i]=len(left)
        for attr in left:
            lhs[attr].append(i)
        for attr in right:
            rhs[i].append(attr)
    
    while todo:
        curA=todo.pop()
        aplus.add(curA)
        for fd in lhs[curA]:
            c[fd]-=1
            if c[fd]==0:
                for newA in rhs[fd]:
                    if newA not in aplus:
                        todo.append(newA)
                        
    return aplus

def main():
    return
    
if __name__=="__main__":
    main()
    



"""
Commandline tool capexplain for the Cape system which explains outliers to aggregation queries.

Installation through setuptools will create a script capexplain. Run "capexplain help" to see usage.
"""

import sys
from enum import Enum, unique
import getopt
import logging
import psycopg2
from inspect import currentframe, getframeinfo
from capexplain.pattern_miner.PatternMiner import PatternFinder, MinerConfig
from capexplain.database.dbaccess import DBConnection
from capexplain.cl.cfgoption import ConfigOpt, OptionType, DictLike
from capexplain.cl.command import CmdTypes, Command, CmdOptions
from capexplain.explain.explanation import ExplanationGenerator, ExplConfig
from capexplain.gui.Cape_GUI import startCapeGUI 
import colorful

# ********************************************************************************
# format for logging
LOGFORMAT='{c.white_on_black}%(levelname)s{c.reset} {c.red}%(asctime)s{c.reset} {c.blue}[%(filename)s:%(funcName)s:%(lineno)d]{c.reset} %(message)s'.format(c=colorful)

# ********************************************************************************
def mineCommand(c,log):
    """
    Command action for mine for patterns (options are parsed commandline options).
    """
    # create configuration based on options
    config=MinerConfig()
    dbconn=DBConnection()

    log.debug("executing mine command")
    c.options.setupConfigAndConnection(dbconn, config)
    
    config.validateConfiguration()
    config.conn = dbconn.connect()
    
    log.debug("connected to database")
    p=PatternFinder(config)
    log.debug("created pattern miner object")
    p.findPattern()
    log.debug("done finding patterns")
    config.conn.close()
    log.debug("closed database connection ... DONE")

# ********************************************************************************
def helpCommand(argv):
    """
    Command for showing help message.
    """
    if len(argv) == 0:
        printHelp()
    else:
        cmdString = argv[0]
        command = next((c for c in COMMANDS if c.cmdstr == cmdString), None)
        if command is None:
            print("unknown command: {}\n\n".format(cmdString))
            printHelp()
        else:
            printHelp(command)

# ********************************************************************************
def statsCommand(command,log):
    """
    Command for printing stats and patterns of previous executions of miner or explainer.
    """
    # create configuration based on options
    config=DictLike()
    dbconn=DBConnection()

    log.debug("executing mine command")

    command.options.setupConfigAndConnection(dbconn, config)
    
    config.validateConfiguration()
    config.conn = dbconn.connect()
    log.debug("connected to database")
    config.conn.close()
    log.debug("closed database connection ... DONE")

# ********************************************************************************
def explainCommand(command,log):
    """
    Command for explaining an outlier.
    """
    # create configuration based on options
    config=ExplConfig()
    dbconn=DBConnection()

    # setup configuration
    command.options.setupConfigAndConnection(dbconn,config)

    # create connection
    config.conn = dbconn.pgconnect()
    config.cur = config.conn.cursor()
    config.pattern_table = dbconn.local_table[:-6]
    log.debug("connected to database")
        
    # do explaining
    log.debug("executing explain command")
    e = ExplanationGenerator(config, None)
    log.debug("created ExplanationGenerator")
    e.doExplain()
    log.debug("explanation generation finished")
    config.conn.close()

# ********************************************************************************
def guiCommand(command,log):
    """
    Command for printing stats and patterns of previous executions of miner or explainer.
    """
    # create configuration based on options
    config=DictLike()
    dbconn=DBConnection()

    log.debug("executing GUI command")

    command.options.setupConfigAndConnection(dbconn, config)
    
    config.validateConfiguration()
    conn=dbconn.pgconnect()
    config.conn = conn
    config.cur = config.conn.cursor()
    log.debug("connected to database")
    #TODO gui=Cape_UI() # (config, dbconn)
    # start the gui gui.start()
    startCapeGUI(conn=conn,config=config)

    conn.close()
    log.debug("closed database connection ... DONE")

    

# ********************************************************************************
# options for different commands using ConfigOpt
COMMON_OPTIONS =  [ ConfigOpt(longopt='log', shortopt='l', desc='select log level {DEBUG,INFO,WARNING,ERROR}', hasarg=True, value="ERROR"),
                    ConfigOpt(longopt='help', desc='show this help message'),
]

DB_OPTIONS = [ ConfigOpt(longopt='host', shortopt='h', desc='database connection host IP address', hasarg=True),
            ConfigOpt(longopt='user', shortopt='u', desc='database connection user', hasarg=True),
            ConfigOpt(longopt='password', shortopt='p', desc='database connection password', hasarg=True),
            ConfigOpt(longopt='db', shortopt='d', desc='database name', hasarg=True),
            ConfigOpt(longopt='port', shortopt='P', desc='database connection port', otype=OptionType.Int, hasarg=True)
]

MINE_OPTIONS = COMMON_OPTIONS + DB_OPTIONS + [
            ConfigOpt(longopt='target-table', shortopt='t', desc='mine patterns for this table', hasarg=True, cfgFieldName='table'),
            ConfigOpt(longopt='gof-const', shortopt=None, desc='goodness-of-fit threshold for constant regression', hasarg=True, otype=OptionType.Float, cfgFieldName='theta_c'),
            ConfigOpt(longopt='gof-linear', shortopt=None, desc='goodness-of-fit threshold for linear regression', hasarg=True, otype=OptionType.Float, cfgFieldName='theta_l'),
            ConfigOpt(longopt='confidence', shortopt=None, desc='global confidence threshold', hasarg=True, otype=OptionType.Float, cfgFieldName='lamb'),
            ConfigOpt(longopt='regpackage', shortopt='r', desc=('regression analysis package to use {}'.format(MinerConfig.STATS_MODELS)), hasarg=True,cfgFieldName='reg_package'),
            ConfigOpt(longopt='local-support', shortopt=None, desc='local support threshold', hasarg=True, otype=OptionType.Int,cfgFieldName='supp_l'),
            ConfigOpt(longopt='global-support', shortopt=None, desc='global support thresh', hasarg=True, otype=OptionType.Int,cfgFieldName='supp_g'),
            ConfigOpt(longopt='fd-optimizations', shortopt='f', desc='activate functional dependency detection and optimizations',cfgFieldName='fd_check'),
            ConfigOpt(longopt='algorithm', shortopt='a', desc='algorithm to use for pattern mining {}'.format(MinerConfig.ALGORITHMS), hasarg=True),
            ConfigOpt(longopt='show-progress', shortopt=None, desc='show progress meters', otype=OptionType.Boolean, hasarg=True, cfgFieldName='showProgress'),
]

# EXPLAIN_OPTIONS = COMMON_OPTIONS + [ ConfigOpt(longopt='qfile', shortopt='q', desc='file storing aggregation query result', hasarg=True, cfgFieldName='query_result_file'),
#                     ConfigOpt(longopt='cfile', shortopt='c', desc='file storing patterns', hasarg=True, cfgFieldName='constraint_file'),
#                     ConfigOpt(longopt='ufile', shortopt='u', desc='file storing user question', hasarg=True, cfgFieldName='user_question_file'),
#                     ConfigOpt(longopt='ofile', shortopt='o', desc='file to write output to', hasarg=True, cfgFieldName='outfile'),
#                     ConfigOpt(longopt='epsilon', shortopt='e', desc='file to write output to', hasarg=True, otype=OptionType.Float, cfgFieldName='constraint_epsilon'),
#                     ConfigOpt(longopt='aggcolumn', shortopt='a', desc='column that was input to the aggregation function', hasarg=True, cfgFieldName='aggregate_column'), 
# ]

EXPLAIN_OPTIONS = COMMON_OPTIONS + DB_OPTIONS + [ 
            ConfigOpt(longopt='ptable', desc='table storing aggregate regression patterns', hasarg=True, cfgFieldName='pattern_table'),
            ConfigOpt(longopt='qtable', desc='table storing aggregation query result', hasarg=True, cfgFieldName='query_result_table'),
            ConfigOpt(longopt='ufile', desc='file storing user question', hasarg=True, cfgFieldName='user_question_file'),
            ConfigOpt(longopt='ofile', shortopt='o', desc='file to write output to', hasarg=True, cfgFieldName='outfile'),
            ConfigOpt(longopt='aggcolumn', shortopt='a', desc='column that was input to the aggregation function', hasarg=True, cfgFieldName='aggregate_column'), 
]

STATS_OPTIONS = COMMON_OPTIONS + [ 
]

HELP_OPTIONS = [ ConfigOpt(longopt='log', shortopt='l', desc='select log level {DEBUG,INFO,WARNING,ERROR}', hasarg=True, value="ERROR"),
]

GUI_OPTIONS = COMMON_OPTIONS + DB_OPTIONS +[

]

# mapping strings to log levels
LOGLEVELS_MAP = { "DEBUG": logging.DEBUG,
                  "INFO": logging.INFO,
                  "WARNING": logging.WARNING,
                  "ERROR": logging.ERROR,
}

# commands supported by Cape
COMMANDS = [ Command(cmd=CmdTypes.Mine,cmdstr='mine',options=CmdOptions(MINE_OPTIONS),helpMessage='Mining patterns that hold for a relation (necessary preprocessing step for generating explanations.',execute=mineCommand),
             Command(cmd=CmdTypes.Explain,cmdstr='explain',options=CmdOptions(EXPLAIN_OPTIONS),helpMessage='Generate explanations for an aggregation result (patterns should have been mined upfront using mine).', execute=explainCommand),
             Command(cmd=CmdTypes.Stats,cmdstr='stats',options=CmdOptions(STATS_OPTIONS),helpMessage='Extracting statistics from database collected during previous mining executions.',execute=statsCommand),
             Command(cmd=CmdTypes.Help,cmdstr='help',options=CmdOptions(HELP_OPTIONS),helpMessage='Show general or command specific help.', execute=helpCommand),
             Command(cmd=CmdTypes.GUI,cmdstr='gui',options=CmdOptions(GUI_OPTIONS),helpMessage='Open the Cape graphical explanation explorer.', execute=guiCommand) 
]

# maps command names to Command objects
COMMAND_BY_TYPE = { x.cmd : x for x in COMMANDS}

# ********************************************************************************
def getCmdList():
    """ 
    Return a help message for the supported commands.
    """
    return "\n".join((x.helpString() for x in COMMANDS))

# ********************************************************************************
def printHelp(c=None):
    """
    Print command specific or general help message.
    """
    if c is None:
        helpMessage = 'capexplain COMMAND [OPTIONS]:\n\texplain unusually high or low aggregation query results.\n\nAVAILABLE COMMANDS:\n\n' + getCmdList()
    elif c.cmd == CmdTypes.Help:
        helpMessage = 'capexplain help [COMMAND]: show general or command specific help message.\n\nAVAILABLE COMMANDS:\n\n' + getCmdList()
    else:
        helpMessage = 'capexplain {} [OPTIONS]:\n\t{}\n\nSUPPORTED OPTIONS:\n{}'.format(
            c.cmdstr,
            c.helpMessage,
            "\n".join(o.helpString() for o in c.options.optionlist))
    print(helpMessage)

def parseOptions(argv):
    """
    Parse options from commandline.
    """
    # detect command
    if len(argv) > 0:
        cmdString = argv[0]
        command = next((c for c in COMMANDS if c.cmdstr == cmdString), None)
        argv=argv[1:]
    else:
        command = None
    if command is None:
        printHelp()
        sys.exit(2)

    if (command.cmd == CmdTypes.Help):
        helpCommand(argv)
        sys.exit(1)
    
    # parse options
    o = command.options
    try:
        opts, args = getopt.getopt(argv, o.shortopts, o.longopts)
    except getopt.GetoptError as e:
        print("Exception {}\n\n{}\n", type(e), e.args)
        printHelp(command)
        sys.exit(2)
        
    # store command options and args
    for opt, arg in opts:
        if opt in o.shortopt_map:
            cmdopt = o.shortopt_map[opt]
            if cmdopt.hasarg:
                cmdopt.value = arg
            else:
                cmdopt.value = True
        elif opt in o.longopt_map:
            cmdopt = o.longopt_map[opt]
            if cmdopt.hasarg:
                cmdopt.value = arg
            else:
                cmdopt.value = True
        else:
            print("invalid option {} for command {}".format(opt, command))
            sys.exit(2)

    # cast options 
    for cmdopt in o.optionlist:
        cmdopt.castValue()

    # print help and exist if user asked for help
    if o.cmdConfig['help'].value is not None:
        printHelp(command)
        sys.exit(2)
        
    # if log level is set then record that
    if "log" in o.cmdConfig:
        loglevel = o.cmdConfig["log"].value
        
    return loglevel, command




def main(argv=sys.argv[1:]):
    """
    cape main function
    """
    loglevel, command = parseOptions(argv)
    logging.basicConfig(level=LOGLEVELS_MAP[loglevel],format=LOGFORMAT)    
    log = logging.getLogger(__name__)
    log.debug("...started")
    log.debug("parsed options")
    log.debug("execute command: {}".format(command.cmdstr))
    command.execute(command,log=log)
    log.debug("Cape is finished")

if __name__=="__main__":
    main(sys.argv[1:])



"""
 Profiling including user controlled times.
"""

from time import time

# ********************************************************************************
class ExecStats:
    """
    Statistics gathered during mining
    """

    TIMERS={'total'    }

    COUNTERS={}
    
    def __init__(self):
        self.time = {}
        self.numcalls = {}
        self.timer = {}
        self.counters = {}
        for t in self.TIMERS:
            self.time[t] = 0
            self.numcalls[t] = 0
            self.timer[t] = 0
        for c in self.COUNTERS:
            self.counters[c] = 0

    def formatStats(self):
        res='TIME MEASUREMENTS:\n{:*<40}\n'.format('')
        for t in sorted(self.time):
            res+="{:40}#calls: {:<20d}total-time: {:.8f}\n".format(t,self.numcalls[t],self.time[t])
        res+='\nCOUNTERS:\n{:*<40}\n'.format('')
        for c in sorted(self.counters):
            res+="number of {:20}{:<20d}\n".format(c,self.counters[c])
        return res

    def startTimer(self, name):
        self.timer[name] = time()
        self.numcalls[name]+=1
        #log.debug("numcalls: %d", self.numcalls[name])

    def stopTimer(self, name):
        measuredTime = time() - self.timer[name]
        #log.debug("timer %s ran for %f secs", name, measuredTime)
        self.time[name] += measuredTime

    def incr(self,name):
        self.counters[name] += 1

    def getCounter(self,name):
        return self.counters[name]
